Package: OdysseusPathwayModule
Title: Cohort Pathway Analysis with Pre-Index Event Support
Version: 0.0.1
Authors@R: 
    person("Alexander", "Alexeyuk",
           email = "AlexanderAlexeyuk@gmail.com", role = c("aut", "cre"))
Maintainer: Alexander Alexeyuk <AlexanderAlexeyuk@gmail.com>
Description: Provides cohort pathway analysis for Observational Medical
    Outcomes Partnership (OMOP) Common Data Model databases,
    including both standard (post-index) and pre-index pathway analyses.
    The pre-index analysis identifies sequences of events occurring in a
    lookback window before the target cohort index date. Built on the
    'CohortPathways' analysis framework originally developed by Christopher
    Knoll and the Observational Health Data Sciences and Informatics
    community through 'WebAPI'. Methodological background and the
    originating implementation are described in
    <https://github.com/OHDSI/CohortPathways>.
License: Apache License (>= 2)
Depends: R (>= 4.1.0)
Encoding: UTF-8
RoxygenNote: 7.3.3
Imports: DatabaseConnector, SqlRender
Suggests: Eunomia, igraph, knitr, rmarkdown, testthat
VignetteBuilder: knitr
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-04-17 08:02:35 UTC; AlexanderAleksiayuk
Author: Alexander Alexeyuk [aut, cre]
Repository: CRAN
Date/Publication: 2026-04-21 19:22:15 UTC
Built: R 4.6.0; ; 2026-04-30 14:42:44 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(plot,event_sequence_graph)
S3method(print,event_sequence_graph)
export(buildEventSequenceGraph)
export(executeCohortPathways)
importFrom(stats,aggregate)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, eval=FALSE--------------------------------------------------------
# library(OdysseusPathwayModule)
# library(Eunomia)
# 
# connectionDetails <- Eunomia::getEunomiaConnectionDetails()

## ----cohorts, eval=FALSE------------------------------------------------------
# generationSet <- Eunomia::createCohorts(connectionDetails)
# 
# generationSet

## ----post_index, eval=FALSE---------------------------------------------------
# postIndexResults <- executeCohortPathways(
#   connectionDetails = connectionDetails,
#   cohortDatabaseSchema = "main",
#   cohortTableName = "cohort",
#   targetCohortIds = 4,
#   eventCohortIds = c(1, 2, 3),
#   maxDepth = 3,
#   collapseWindow = 30
# )

## ----post_index_names, eval=FALSE---------------------------------------------
# names(postIndexResults)

## ----post_index_inspect, eval=FALSE-------------------------------------------
# head(postIndexResults$pathwaysAnalysisPathsData)
# head(postIndexResults$pathwayAnalysisCodesLong)

## ----pre_index, eval=FALSE----------------------------------------------------
# preIndexResults <- executeCohortPathways(
#   connectionDetails = connectionDetails,
#   cohortDatabaseSchema = "main",
#   cohortTableName = "cohort",
#   targetCohortIds = 4,
#   eventCohortIds = c(1, 2, 3),
#   analysisType = "pre-index",
#   lookbackStartDay = -365,
#   lookbackEndDay = -1,
#   maxDepth = 3,
#   collapseWindow = 30
# )

## ----pre_index_90, eval=FALSE-------------------------------------------------
# preIndex90 <- executeCohortPathways(
#   connectionDetails = connectionDetails,
#   cohortDatabaseSchema = "main",
#   cohortTableName = "cohort",
#   targetCohortIds = 4,
#   eventCohortIds = c(1, 2, 3),
#   analysisType = "pre-index",
#   lookbackStartDay = -90,
#   lookbackEndDay = -1,
#   maxDepth = 3
# )

## ----code_mapping, eval=FALSE-------------------------------------------------
# subset(
#   postIndexResults$pathwayAnalysisCodesLong,
#   select = c(pathwayAnalysisGenerationId, code, targetCohortId, eventCohortId, isCombo, numberOfEvents)
# )

## ----separate_tables, eval=FALSE----------------------------------------------
# connection <- DatabaseConnector::connect(connectionDetails)
# 
# DatabaseConnector::executeSql(connection, "DROP TABLE IF EXISTS target_cohorts;")
# DatabaseConnector::executeSql(connection, "DROP TABLE IF EXISTS event_cohorts;")
# 
# DatabaseConnector::executeSql(
#   connection,
#   "CREATE TABLE target_cohorts AS
#      SELECT *
#      FROM main.cohort
#      WHERE cohort_definition_id = 4;"
# )
# 
# DatabaseConnector::executeSql(
#   connection,
#   "CREATE TABLE event_cohorts AS
#      SELECT *
#      FROM main.cohort
#      WHERE cohort_definition_id IN (1, 2, 3);"
# )
# 
# resultsSeparateTables <- executeCohortPathways(
#   connectionDetails = connectionDetails,
#   cohortDatabaseSchema = "main",
#   cohortTableName = "target_cohorts",
#   outcomeDatabaseSchema = "main",
#   outcomeTableName = "event_cohorts",
#   targetCohortIds = 4,
#   eventCohortIds = c(1, 2, 3)
# )
# 
# DatabaseConnector::disconnect(connection)

## ----esg_mock_data, eval=FALSE------------------------------------------------
# # --- Mock cpResults with multi-step pathways ---
# # Bitmask combo codes: 2 = Celecoxib, 4 = Diclofenac, 8 = GiBleed
# mockPathsData <- data.frame(
#   pathwayAnalysisGenerationId = rep(1L, 5),
#   targetCohortId              = rep(4L, 5),
#   step1      = c( 2L,  2L,  4L,  4L,  2L),
#   step2      = c( 4L,  8L,  2L,  8L, NA),
#   step3      = c( 8L, NA,   8L, NA,  NA),
#   countValue = c(120L, 80L, 95L, 65L, 40L)
# )
# 
# mockCodesLong <- data.frame(
#   pathwayAnalysisGenerationId = rep(1L, 3),
#   code           = c(2L, 4L, 8L),
#   targetCohortId = rep(4L, 3),
#   eventCohortId  = c(1L, 2L, 3L),
#   isCombo        = rep(0L, 3),
#   numberOfEvents = rep(1L, 3)
# )
# 
# mockIsCombo <- data.frame(
#   targetCohortId = rep(4L, 3),
#   comboId        = c(2L, 4L, 8L),
#   numberOfEvents = rep(1L, 3),
#   isCombo        = rep(0L, 3)
# )
# 
# mockCpResults <- list(
#   pathwayAnalysisStatsData   = data.frame(
#     pathwayAnalysisGenerationId = 1L,
#     targetCohortId = 4L,
#     countValue = 400L
#   ),
#   pathwaysAnalysisPathsData  = mockPathsData,
#   pathwaysAnalysisEventsData = data.frame(eventCohortId = 1:3, countValue = c(240L, 215L, 360L)),
#   pathwaycomboIds            = data.frame(comboIds = c(2L, 4L, 8L)),
#   pathwayAnalysisCodesLong   = mockCodesLong,
#   isCombo                    = mockIsCombo,
#   pathwayAnalysisCodesData   = data.frame(
#     pathwayAnalysisGenerationId = rep(1L, 3),
#     code    = c(2L, 4L, 8L),
#     isCombo = rep(0L, 3)
#   )
# )

## ----event_sequence_graph, eval=FALSE-----------------------------------------
# # Map cohort IDs to human-readable names
# generationSet <- data.frame(
#   cohortId   = c(1L, 2L, 3L, 4L),
#   cohortName = c("Celecoxib", "Diclofenac", "GiBleed", "NSAIDs")
# )
# 
# esg <- buildEventSequenceGraph(
#   cpResults     = mockCpResults,
#   generationSet = generationSet,
#   maxSteps      = 3,
#   minCount      = 1
# )
# 
# # Print a summary
# esg

## ----esg_real_note, eval=FALSE------------------------------------------------
# # With real data:
# # generationSet <- Eunomia::createCohorts(connectionDetails)
# # generationSet$cohortName <- generationSet$name
# #
# # esg <- buildEventSequenceGraph(
# #   cpResults     = postIndexResults,
# #   generationSet = generationSet,
# #   maxSteps      = 3,
# #   minCount      = 5
# # )

## ----graph_components, eval=FALSE---------------------------------------------
# # The igraph object — vertices are (event, step) pairs, edges are transitions
# ig <- esg$graph
# 
# # Vertex attributes
# igraph::V(ig)$eventName   # human-readable event names
# igraph::V(ig)$step        # pathway step number
# igraph::V(ig)$count       # patient count at this node
# igraph::V(ig)$share       # share within the step (sums to 1)
# 
# # Edge attributes
# igraph::E(ig)$weight      # patient count crossing this transition
# igraph::E(ig)$probability # transition probability (sums to 1 per source)
# igraph::E(ig)$sourceStep
# igraph::E(ig)$targetStep
# 
# # Decoded pathways
# head(esg$sequences)
# 
# # Summary statistics
# esg$summary

## ----plot_esg, eval=FALSE-----------------------------------------------------
# # Default plot
# plot(esg)
# 
# # Customized plot
# plot(esg,
#   colorPalette    = c("#1b9e77", "#d95f02", "#7570b3"),
#   edgeWidthRange  = c(1, 10),
#   vertexSizeRange = c(10, 30),
#   main            = "Post-Index Treatment Pathways"
# )

## ----transition_probs, eval=FALSE---------------------------------------------
# # Extract edge data frame
# edgeDf <- igraph::as_data_frame(esg$graph, what = "edges")
# 
# # View transitions from Step 1 to Step 2
# edgeDf[edgeDf$sourceStep == 1, ]

## ----igraph_analysis, eval=FALSE----------------------------------------------
# ig <- esg$graph
# 
# # Out-degree: how many distinct next-step events each node leads to
# igraph::degree(ig, mode = "out")
# 
# # Weighted betweenness (inverse weight = lower traffic → higher betweenness)
# igraph::betweenness(ig, weights = 1 / igraph::E(ig)$weight)
# 
# # Shortest weighted paths between all pairs
# igraph::distances(ig, weights = 1 / igraph::E(ig)$weight)
# 
# # Identify hubs and authorities (HITS)
# igraph::hub_score(ig, weights = igraph::E(ig)$weight)$vector
# igraph::authority_score(ig, weights = igraph::E(ig)$weight)$vector
# 
# # Export to data frames for use outside igraph
# vertDf <- igraph::as_data_frame(ig, what = "vertices")
# edgeDf <- igraph::as_data_frame(ig, what = "edges")

# Copyright 2026 Odysseus Data Services
#
# This file is part of OdysseusPathwayModule
#
# Adapted from CohortPathways by OHDSI (https://github.com/OHDSI/CohortPathways).
# The core implementation of CohortPathway analysis was developed by Christopher
# Knoll (github: chrisknoll), and enhanced by members of the OHDSI community:
# mick-iqvia (github), Odysseus Inc., and many other collaborators on the
# OHDSI/WebAPI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


#' Execute cohort pathway analysis.
#'
#' @description
#' Runs the cohort pathways on all instantiated combinations of target and event cohorts.
#' Assumes the cohorts have already been instantiated.
#'
#' Supports two analysis types:
#' \itemize{
#'   \item \code{"post-index"} (default): events occurring **after** the target cohort
#'         index date (within the target cohort period).
#'   \item \code{"pre-index"}: events occurring **before** the target cohort index date,
#'         in a configurable lookback window. For example, what events occurred in
#'         the 365 days before cohort entry.
#' }
#'
#' Target and event cohorts can reside in different schemas/tables:
#' \itemize{
#'   \item \code{cohortDatabaseSchema} / \code{cohortTableName} for target cohorts.
#'   \item \code{outcomeDatabaseSchema} / \code{outcomeTableName} for event (outcome) cohorts.
#'         If not specified, defaults to the same as the target.
#' }
#'
#' @param connectionDetails   An object of type connectionDetails as created using the
#'                            \code{\link[DatabaseConnector]{createConnectionDetails}} function in the
#'                            DatabaseConnector package. Can be left NULL if \code{connection} is
#'                            provided.
#' @param connection          An object of type \code{connection} as created using the
#'                            \code{\link[DatabaseConnector]{connect}} function in the
#'                            DatabaseConnector package. Can be left NULL if connectionDetails
#'                            is provided, in which case a new connection will be opened at the start
#'                            of the function, and closed when the function finishes.
#' @param cohortDatabaseSchema   Schema name where your **target** cohort table resides.
#'                               For SQL Server this should include both database and schema,
#'                               e.g. 'scratch.dbo'.
#' @param cohortTableName        The name of the **target** cohort table (default: "cohort").
#' @param outcomeDatabaseSchema  Schema name where your **event/outcome** cohort table resides.
#'                               Defaults to \code{cohortDatabaseSchema} if not specified.
#' @param outcomeTableName       The name of the **event/outcome** cohort table.
#'                               Defaults to \code{cohortTableName} if not specified.
#' @param targetCohortIds     A vector of one or more Cohort Ids corresponding to target cohort(s).
#' @param eventCohortIds      A vector of one or more Cohort Ids corresponding to event cohort(s).
#' @param analysisType        Character. Either \code{"post-index"} (default) to analyze events
#'                            after the index date, or \code{"pre-index"} to analyze events
#'                            before the index date.
#' @param lookbackStartDay    (Only used when \code{analysisType = "pre-index"}.)
#'                            Start of the lookback window in days relative to the index date
#'                            (default: -365). Must be negative.
#' @param lookbackEndDay      (Only used when \code{analysisType = "pre-index"}.)
#'                            End of the lookback window in days relative to the index date
#'                            (default: -1). Must be negative.
#' @param tempEmulationSchema   Some database platforms like Oracle and Impala do not truly support
#'                              temp tables. To emulate temp tables, provide a schema with write
#'                              privileges where temp tables can be created.
#' @param minCellCount          (Default = 5) The minimum cell count for fields containing person
#'                              counts or fractions.
#' @param allowRepeats          (Default = FALSE) Allow cohort events/combos to appear multiple
#'                              times in the same pathway.
#' @param maxDepth              (Default = 5) Maximum number of steps in a given pathway.
#' @param collapseWindow        (Default = 30) Any dates found within the specified collapse days
#'                              will be reassigned the earliest date.
#' @return                      A list of data frames containing pathway analysis results.
#'
#' @examples
#' \donttest{
#' if (requireNamespace("Eunomia", quietly = TRUE) &&
#'     exists("createCohorts", where = asNamespace("Eunomia"), inherits = FALSE)) {
#'   connectionDetails <- Eunomia::getEunomiaConnectionDetails()
#'   createCohorts <- get("createCohorts", envir = asNamespace("Eunomia"))
#'   createCohorts(connectionDetails)
#'
#'   postIndexResults <- executeCohortPathways(
#'     connectionDetails = connectionDetails,
#'     cohortDatabaseSchema = "main",
#'     cohortTableName = "cohort",
#'     targetCohortIds = 4,
#'     eventCohortIds = c(1, 2, 3),
#'     maxDepth = 3
#'   )
#'
#'   preIndexResults <- executeCohortPathways(
#'     connectionDetails = connectionDetails,
#'     cohortDatabaseSchema = "main",
#'     cohortTableName = "cohort",
#'     targetCohortIds = 4,
#'     eventCohortIds = c(1, 2, 3),
#'     analysisType = "pre-index",
#'     lookbackStartDay = -365,
#'     lookbackEndDay = -1,
#'     maxDepth = 3
#'   )
#'
#'   names(postIndexResults)
#'   names(preIndexResults)
#' }
#' }
#'
#' @export
executeCohortPathways <- function(connectionDetails = NULL,
                                  connection = NULL,
                                  cohortDatabaseSchema,
                                  cohortTableName = "cohort",
                                  outcomeDatabaseSchema = cohortDatabaseSchema,
                                  outcomeTableName = cohortTableName,
                                  tempEmulationSchema = getOption("sqlRenderTempEmulationSchema"),
                                  targetCohortIds,
                                  eventCohortIds,
                                  analysisType = "post-index",
                                  lookbackStartDay = -365,
                                  lookbackEndDay = -1,
                                  minCellCount = 5,
                                  allowRepeats = FALSE,
                                  maxDepth = 5,
                                  collapseWindow = 30) {
  start <- Sys.time()

  # --- Input validation (base R) ---
  analysisType <- match.arg(analysisType, c("post-index", "pre-index"))

  if (!is.character(cohortDatabaseSchema) || length(cohortDatabaseSchema) < 1) {
    stop("'cohortDatabaseSchema' must be a non-empty character string.")
  }
  if (!is.character(outcomeDatabaseSchema) || length(outcomeDatabaseSchema) < 1) {
    stop("'outcomeDatabaseSchema' must be a non-empty character string.")
  }
  if (!is.numeric(minCellCount) || length(minCellCount) != 1 || minCellCount < 0) {
    stop("'minCellCount' must be a single non-negative number.")
  }
  if (!is.logical(allowRepeats) || length(allowRepeats) != 1) {
    stop("'allowRepeats' must be a single logical value.")
  }
  if (!is.numeric(maxDepth) || length(maxDepth) != 1 || maxDepth < 1) {
    stop("'maxDepth' must be a single positive number.")
  }
  if (!is.numeric(collapseWindow) || length(collapseWindow) != 1 || collapseWindow < 0) {
    stop("'collapseWindow' must be a single non-negative number.")
  }

  if (analysisType == "pre-index") {
    if (!is.numeric(lookbackStartDay) || length(lookbackStartDay) != 1 || lookbackStartDay >= 0) {
      stop("'lookbackStartDay' must be a single negative number.")
    }
    if (!is.numeric(lookbackEndDay) || length(lookbackEndDay) != 1 || lookbackEndDay >= 0) {
      stop("'lookbackEndDay' must be a single negative number.")
    }
    if (lookbackStartDay >= lookbackEndDay) {
      stop(
        paste0(
          "'lookbackStartDay' (", lookbackStartDay,
          ") must be less than 'lookbackEndDay' (", lookbackEndDay, ").",
          " Example: lookbackStartDay = -365, lookbackEndDay = -1."
        )
      )
    }
  }

  message(paste0("Run Cohort Pathways (", analysisType, ") started at ", start))

  if (analysisType == "pre-index") {
    message(
      sprintf(
        "   Pre-index lookback window: [%d, %d] days relative to index date.",
        lookbackStartDay, lookbackEndDay
      )
    )
  }

  # allow repeats is used as text 'true' or 'false' in sql
  allowRepeatsStr <- if (allowRepeats) "true" else "false"

  if (is.null(connection)) {
    connection <- DatabaseConnector::connect(connectionDetails)
    on.exit(DatabaseConnector::disconnect(connection))
  }

  # --- Verify target cohort table exists ---
  cohortTableName <- tolower(cohortTableName)
  tablesInCohortSchema <-
    DatabaseConnector::getTableNames(
      connection = connection,
      databaseSchema = cohortDatabaseSchema
    ) |>
    tolower()

  if (!cohortTableName %in% c(tablesInCohortSchema, "")) {
    stop(
      paste0(
        "Target cohort table '", toupper(cohortTableName),
        "' not found in schema '", cohortDatabaseSchema, "'"
      )
    )
  }

  # --- Verify event/outcome cohort table exists ---
  outcomeTableName <- tolower(outcomeTableName)
  tablesInOutcomeSchema <-
    DatabaseConnector::getTableNames(
      connection = connection,
      databaseSchema = outcomeDatabaseSchema
    ) |>
    tolower()

  if (!outcomeTableName %in% c(tablesInOutcomeSchema, "")) {
    stop(
      paste0(
        "Event/outcome cohort table '", toupper(outcomeTableName),
        "' not found in schema '", outcomeDatabaseSchema, "'"
      )
    )
  }

  # --- Count cohorts ---
  # Target cohort counts
  targetCounts <- DatabaseConnector::renderTranslateQuerySql(
    connection = connection,
    sql = "SELECT cohort_definition_id AS cohort_id,
              COUNT(*) AS cohort_entries,
              COUNT(DISTINCT subject_id) AS cohort_subjects
          FROM @cohort_database_schema.@cohort_table
          {@cohort_ids != ''} ? {WHERE cohort_definition_id IN (@cohort_ids)}
          GROUP BY cohort_definition_id;",
    cohort_database_schema = cohortDatabaseSchema,
    cohort_table = cohortTableName,
    cohort_ids = targetCohortIds,
    snakeCaseToCamelCase = TRUE
  )

  # Event cohort counts
  eventCounts <- DatabaseConnector::renderTranslateQuerySql(
    connection = connection,
    sql = "SELECT cohort_definition_id AS cohort_id,
              COUNT(*) AS cohort_entries,
              COUNT(DISTINCT subject_id) AS cohort_subjects
          FROM @cohort_database_schema.@cohort_table
          {@cohort_ids != ''} ? {WHERE cohort_definition_id IN (@cohort_ids)}
          GROUP BY cohort_definition_id;",
    cohort_database_schema = outcomeDatabaseSchema,
    cohort_table = outcomeTableName,
    cohort_ids = eventCohortIds,
    snakeCaseToCamelCase = TRUE
  )

  if (nrow(targetCounts) == 0 ||
      !any(targetCounts$cohortId %in% targetCohortIds)) {
    stop("None of the target cohorts are instantiated.")
  }

  if (nrow(eventCounts) == 0 ||
      !any(eventCounts$cohortId %in% eventCohortIds)) {
    stop("None of the event cohorts are instantiated.")
  }

  nTargetFound <- sum(targetCounts$cohortId %in% targetCohortIds)
  nEventFound <- sum(eventCounts$cohortId %in% eventCohortIds)

  if (nTargetFound < length(targetCohortIds) ||
      nEventFound < length(eventCohortIds)) {
    message("Not all cohorts have more than 0 records.")
    message(
      sprintf("    Found %s of %s target cohorts instantiated.", nTargetFound, length(targetCohortIds))
    )
    message(
      sprintf("    Found %s of %s event cohorts instantiated.", nEventFound, length(eventCohortIds))
    )
  }

  targetCohortTable <- paste0(cohortDatabaseSchema, ".", cohortTableName)
  eventCohortTable <- paste0(outcomeDatabaseSchema, ".", outcomeTableName)

  instantiatedEventCohortIds <-
    intersect(x = eventCohortIds, y = eventCounts$cohortId)
  instantiatedTargetCohortIds <-
    intersect(x = targetCohortIds, y = targetCounts$cohortId)

  # --- Select SQL based on analysis type ---
  sqlFile <- if (analysisType == "pre-index") {
    "RunPreIndexPathwayAnalysis.sql"
  } else {
    "RunPathwayAnalysis.sql"
  }

  pathwayAnalysisSql <-
    SqlRender::readSql(
      sourceFile = system.file(
        "sql", "sql_server", sqlFile,
        package = utils::packageName()
      )
    )

  generationToTarget <- data.frame(
    pathwayAnalysisGenerationId = integer(),
    targetCohortId = integer(),
    stringsAsFactors = FALSE
  )
  eventCohortIdIndexMaps <- data.frame(
    eventCohortId = sort(unique(instantiatedEventCohortIds)),
    stringsAsFactors = FALSE
  )
  eventCohortIdIndexMaps$cohortIndex <- seq_len(nrow(eventCohortIdIndexMaps))

  pathwayAnalysisStatsData <- list()
  pathwaysAnalysisPathsData <- list()
  pathwaysAnalysisEventsData <- list()

  for (i in seq_along(instantiatedTargetCohortIds)) {
    targetCohortId <- instantiatedTargetCohortIds[[i]]

    generationId <-
      (as.integer(format(Sys.Date(), "%Y%m%d")) * 1000) +
      sample(x = 1:1000, size = 1, replace = FALSE)

    eventCohortIdIndexMap <- paste0(
      "SELECT ", eventCohortIdIndexMaps$eventCohortId,
      " AS cohort_definition_id, ", eventCohortIdIndexMaps$cohortIndex,
      " AS cohort_index",
      collapse = " union all "
    )

    message(
      paste0(
        "   Generating ", analysisType, " pathways for target cohort: ",
        targetCohortId,
        ". Generation id: ", generationId, "."
      )
    )

    # Build SQL parameters
    sqlParams <- list(
      connection = connection,
      sql = pathwayAnalysisSql,
      profile = FALSE,
      progressBar = TRUE,
      reportOverallTime = FALSE,
      tempEmulationSchema = tempEmulationSchema,
      allow_repeats = allowRepeatsStr,
      combo_window = collapseWindow,
      max_depth = maxDepth,
      pathway_target_cohort_id = targetCohortId,
      target_cohort_table = targetCohortTable,
      event_cohort_table = eventCohortTable,
      generation_id = generationId,
      event_cohort_id_index_map = eventCohortIdIndexMap
    )

    # Add lookback parameters for pre-index analysis
    if (analysisType == "pre-index") {
      sqlParams$lookback_start_day <- lookbackStartDay
      sqlParams$lookback_end_day <- lookbackEndDay
    }

    do.call(DatabaseConnector::renderTranslateExecuteSql, sqlParams)

    pathwayAnalysisStatsData[[i]] <-
      DatabaseConnector::renderTranslateQuerySql(
        connection = connection,
        sql = "SELECT * FROM #pa_stats;",
        snakeCaseToCamelCase = TRUE
      )
    if (!"pathwayAnalysisGenerationId" %in% names(pathwayAnalysisStatsData[[i]])) {
      pathwayAnalysisStatsData[[i]]$pathwayAnalysisGenerationId <- rep(generationId, nrow(pathwayAnalysisStatsData[[i]]))
    }
    if (!"targetCohortId" %in% names(pathwayAnalysisStatsData[[i]])) {
      pathwayAnalysisStatsData[[i]]$targetCohortId <- rep(targetCohortId, nrow(pathwayAnalysisStatsData[[i]]))
    }

    pathwaysAnalysisPathsData[[i]] <-
      DatabaseConnector::renderTranslateQuerySql(
        connection = connection,
        sql = "SELECT * FROM #pa_paths;",
        snakeCaseToCamelCase = TRUE
      )
    if (!"pathwayAnalysisGenerationId" %in% names(pathwaysAnalysisPathsData[[i]])) {
      pathwaysAnalysisPathsData[[i]]$pathwayAnalysisGenerationId <- rep(generationId, nrow(pathwaysAnalysisPathsData[[i]]))
    }
    if (!"targetCohortId" %in% names(pathwaysAnalysisPathsData[[i]])) {
      pathwaysAnalysisPathsData[[i]]$targetCohortId <- rep(targetCohortId, nrow(pathwaysAnalysisPathsData[[i]]))
    }

    pathwaysAnalysisEventsData[[i]] <-
      DatabaseConnector::renderTranslateQuerySql(
        connection = connection,
        sql = "SELECT * FROM #pa_events;",
        snakeCaseToCamelCase = TRUE
      )
    if (!"pathwayAnalysisGenerationId" %in% names(pathwaysAnalysisEventsData[[i]])) {
      pathwaysAnalysisEventsData[[i]]$pathwayAnalysisGenerationId <- rep(generationId, nrow(pathwaysAnalysisEventsData[[i]]))
    }
    if (!"targetCohortId" %in% names(pathwaysAnalysisEventsData[[i]])) {
      pathwaysAnalysisEventsData[[i]]$targetCohortId <- rep(targetCohortId, nrow(pathwaysAnalysisEventsData[[i]]))
    }

    DatabaseConnector::renderTranslateExecuteSql(
      connection = connection,
      sql = "DROP TABLE IF EXISTS #pa_paths;
             DROP TABLE IF EXISTS #pa_stats;
             DROP TABLE IF EXISTS #pa_events;",
      profile = FALSE,
      progressBar = TRUE,
      reportOverallTime = FALSE,
      tempEmulationSchema = tempEmulationSchema
    )

    generationToTarget <- rbind(
      generationToTarget,
      data.frame(
        pathwayAnalysisGenerationId = generationId,
        targetCohortId = targetCohortId,
        stringsAsFactors = FALSE
      )
    )
  }

  pathwayAnalysisStatsData <- do.call(rbind, pathwayAnalysisStatsData)
  pathwaysAnalysisPathsData <- do.call(rbind, pathwaysAnalysisPathsData)
  pathwaysAnalysisEventsData <- do.call(rbind, pathwaysAnalysisEventsData)

  stepCols <- grep("^step", names(pathwaysAnalysisPathsData), value = TRUE)
  comboIdsAll <- unlist(pathwaysAnalysisPathsData[, stepCols], use.names = FALSE)
  comboIdsVec <- sort(unique(comboIdsAll[!is.na(comboIdsAll) & comboIdsAll > 0]))
  pathwaycomboIds <- data.frame(comboIds = comboIdsVec)

  if (nrow(pathwaysAnalysisPathsData) > 0 && length(stepCols) > 0) {
    comboByGeneration <- do.call(
      rbind,
      lapply(stepCols, function(stepCol) {
        data.frame(
          pathwayAnalysisGenerationId = pathwaysAnalysisPathsData$pathwayAnalysisGenerationId,
          comboId = pathwaysAnalysisPathsData[[stepCol]],
          stringsAsFactors = FALSE
        )
      })
    )
    comboByGeneration <- comboByGeneration[
      !is.na(comboByGeneration$comboId) & comboByGeneration$comboId > 0,
      c("pathwayAnalysisGenerationId", "comboId")
    ]
    comboByGeneration <- unique(comboByGeneration)
    comboByGeneration <- merge(
      comboByGeneration,
      generationToTarget,
      by = "pathwayAnalysisGenerationId"
    )
  } else {
    comboByGeneration <- data.frame(
      pathwayAnalysisGenerationId = integer(),
      comboId = numeric(),
      targetCohortId = integer(),
      stringsAsFactors = FALSE
    )
  }

  if (nrow(pathwaycomboIds) == 0) {
    pathwayAnalysisCodesLong <- data.frame(
      pathwayAnalysisGenerationId = integer(),
      code = numeric(),
      targetCohortId = integer(),
      eventCohortId = integer(),
      isCombo = integer(),
      numberOfEvents = integer()
    )
    isCombo <- data.frame(
      targetCohortId = integer(),
      comboId = numeric(),
      numberOfEvents = integer(),
      isCombo = integer()
    )
    pathwayAnalysisCodesData <- data.frame(
      pathwayAnalysisGenerationId = integer(),
      code = numeric(),
      isCombo = integer()
    )
  } else {
    pathwayAnalysisCodesLong <- NULL
    for (i in seq_len(nrow(comboByGeneration))) {
      cohortIndex <- extractBitSum(x = comboByGeneration$comboId[i])
      combisData <- data.frame(cohortIndex = cohortIndex, stringsAsFactors = FALSE)
      combisData$comboId <- comboByGeneration$comboId[i]
      combisData$targetCohortId <- comboByGeneration$targetCohortId[i]
      combisData$pathwayAnalysisGenerationId <-
        comboByGeneration$pathwayAnalysisGenerationId[i]
      combisData <- merge(combisData, eventCohortIdIndexMaps, by = "cohortIndex")
      pathwayAnalysisCodesLong <- rbind(combisData, pathwayAnalysisCodesLong)
    }

    codesLongSub <- unique(
      pathwayAnalysisCodesLong[, c("targetCohortId", "comboId", "eventCohortId")]
    )
    isCombo <- aggregate(
      eventCohortId ~ targetCohortId + comboId,
      data = codesLongSub,
      FUN = length
    )
    names(isCombo)[names(isCombo) == "eventCohortId"] <- "numberOfEvents"
    isCombo$isCombo <- ifelse(isCombo$numberOfEvents > 1, 1L, 0L)

    pathwayAnalysisCodesLong <- merge(
      pathwayAnalysisCodesLong, isCombo,
      by = c("targetCohortId", "comboId")
    )
    pathwayAnalysisCodesLong <- pathwayAnalysisCodesLong[, c(
      "pathwayAnalysisGenerationId", "comboId", "targetCohortId",
      "eventCohortId", "isCombo", "numberOfEvents"
    )]
    names(pathwayAnalysisCodesLong)[names(pathwayAnalysisCodesLong) == "comboId"] <- "code"

    pathwayAnalysisCodesData <- unique(
      pathwayAnalysisCodesLong[, c("pathwayAnalysisGenerationId", "code", "isCombo")]
    )
  }

  allData <- list(
    pathwayAnalysisStatsData = pathwayAnalysisStatsData,
    pathwaysAnalysisPathsData = pathwaysAnalysisPathsData,
    pathwaysAnalysisEventsData = pathwaysAnalysisEventsData,
    pathwaycomboIds = pathwaycomboIds,
    pathwayAnalysisCodesLong = pathwayAnalysisCodesLong,
    isCombo = isCombo,
    pathwayAnalysisCodesData = pathwayAnalysisCodesData
  )

  delta <- Sys.time() - start
  message(
    "Computing Cohort Pathways took ",
    signif(delta, 3), " ", attr(delta, "units")
  )
  return(allData)
}

mock_dbconnector <- function(mocks, code) {
  ns <- asNamespace("DatabaseConnector")
  original <- list()

  for (name in names(mocks)) {
    original[[name]] <- get(name, envir = ns)
    unlockBinding(name, ns)
    assign(name, mocks[[name]], envir = ns)
    lockBinding(name, ns)
  }

  on.exit({
    for (name in names(original)) {
      unlockBinding(name, ns)
      assign(name, original[[name]], envir = ns)
      lockBinding(name, ns)
    }
  }, add = TRUE)

  force(code)
}


test_that("executeCohortPathways runs full post-index flow with mocked DatabaseConnector", {
  query_mock <- function(connection, sql, cohort_ids = NULL, ...) {
    if (grepl("cohort_definition_id AS cohort_id", sql, fixed = TRUE)) {
      return(data.frame(
        cohortId = as.integer(cohort_ids),
        cohortEntries = rep(10L, length(cohort_ids)),
        cohortSubjects = rep(9L, length(cohort_ids))
      ))
    }

    if (grepl("SELECT * FROM #pa_stats", sql, fixed = TRUE)) {
      return(data.frame(
        pathwayAnalysisGenerationId = 1L,
        targetCohortId = 10L,
        countValue = 5L
      ))
    }

    if (grepl("SELECT * FROM #pa_paths", sql, fixed = TRUE)) {
      return(data.frame(
        step1 = c(2L, 6L),
        step2 = c(0L, 0L),
        countValue = c(5L, 3L)
      ))
    }

    if (grepl("SELECT * FROM #pa_events", sql, fixed = TRUE)) {
      return(data.frame(eventCohortId = c(1L, 2L), countValue = c(5L, 3L)))
    }

    stop("Unexpected SQL query in mock.")
  }

  execute_calls <- list()
  exec_mock <- function(...) {
    execute_calls[[length(execute_calls) + 1]] <<- list(...)
    invisible(NULL)
  }

  get_tables_mock <- function(...) c("cohort")

  result <- mock_dbconnector(
    mocks = list(
      getTableNames = get_tables_mock,
      renderTranslateQuerySql = query_mock,
      renderTranslateExecuteSql = exec_mock
    ),
    code = OdysseusPathwayModule::executeCohortPathways(
      connection = structure(list(), class = "mock_connection"),
      cohortDatabaseSchema = "main",
      cohortTableName = "cohort",
      outcomeDatabaseSchema = "main",
      outcomeTableName = "cohort",
      targetCohortIds = c(10L),
      eventCohortIds = c(1L, 2L),
      analysisType = "post-index",
      maxDepth = 2,
      collapseWindow = 30
    )
  )

  expect_type(result, "list")
  expect_named(result, c(
    "pathwayAnalysisStatsData",
    "pathwaysAnalysisPathsData",
    "pathwaysAnalysisEventsData",
    "pathwaycomboIds",
    "pathwayAnalysisCodesLong",
    "isCombo",
    "pathwayAnalysisCodesData"
  ))

  expect_gt(nrow(result$pathwaysAnalysisPathsData), 0)
  expect_gt(nrow(result$pathwaycomboIds), 0)
  expect_true(any(result$pathwayAnalysisCodesLong$code > 0))

  # At least one main execution call plus one cleanup call.
  expect_gte(length(execute_calls), 2)
})


test_that("executeCohortPathways passes pre-index lookback params to SQL execution", {
  query_mock <- function(connection, sql, cohort_ids = NULL, ...) {
    if (grepl("cohort_definition_id AS cohort_id", sql, fixed = TRUE)) {
      return(data.frame(
        cohortId = as.integer(cohort_ids),
        cohortEntries = rep(10L, length(cohort_ids)),
        cohortSubjects = rep(9L, length(cohort_ids))
      ))
    }

    if (grepl("SELECT * FROM #pa_stats", sql, fixed = TRUE)) {
      return(data.frame(pathwayAnalysisGenerationId = 1L, targetCohortId = 10L, countValue = 5L))
    }

    if (grepl("SELECT * FROM #pa_paths", sql, fixed = TRUE)) {
      return(data.frame(step1 = 2L, countValue = 5L))
    }

    if (grepl("SELECT * FROM #pa_events", sql, fixed = TRUE)) {
      return(data.frame(eventCohortId = 1L, countValue = 5L))
    }

    stop("Unexpected SQL query in mock.")
  }

  execute_calls <- list()
  exec_mock <- function(...) {
    execute_calls[[length(execute_calls) + 1]] <<- list(...)
    invisible(NULL)
  }

  get_tables_mock <- function(...) c("cohort")

  mock_dbconnector(
    mocks = list(
      getTableNames = get_tables_mock,
      renderTranslateQuerySql = query_mock,
      renderTranslateExecuteSql = exec_mock
    ),
    code = OdysseusPathwayModule::executeCohortPathways(
      connection = structure(list(), class = "mock_connection"),
      cohortDatabaseSchema = "main",
      cohortTableName = "cohort",
      outcomeDatabaseSchema = "main",
      outcomeTableName = "cohort",
      targetCohortIds = 10L,
      eventCohortIds = c(1L, 2L),
      analysisType = "pre-index",
      lookbackStartDay = -90,
      lookbackEndDay = -1,
      maxDepth = 2,
      collapseWindow = 30
    )
  )

  non_cleanup <- execute_calls[vapply(
    execute_calls,
    function(x) is.null(x$lookback_start_day) == FALSE,
    logical(1)
  )]

  expect_gte(length(non_cleanup), 1)
  expect_equal(non_cleanup[[1]]$lookback_start_day, -90)
  expect_equal(non_cleanup[[1]]$lookback_end_day, -1)
})


test_that("executeCohortPathways errors when cohort table is not present", {
  get_tables_mock <- function(...) c("other_table")

  query_mock <- function(...) {
    stop("Should not query counts if table lookup fails")
  }

  exec_mock <- function(...) {
    invisible(NULL)
  }

  expect_error(
    mock_dbconnector(
      mocks = list(
        getTableNames = get_tables_mock,
        renderTranslateQuerySql = query_mock,
        renderTranslateExecuteSql = exec_mock
      ),
      code = OdysseusPathwayModule::executeCohortPathways(
        connection = structure(list(), class = "mock_connection"),
        cohortDatabaseSchema = "main",
        cohortTableName = "cohort",
        targetCohortIds = 10L,
        eventCohortIds = 1L
      )
    ),
    "Target cohort table"
  )
})

test_that("executeCohortPathways returns empty code tables when no positive combos are present", {
  query_mock <- function(connection, sql, cohort_ids = NULL, ...) {
    if (grepl("cohort_definition_id AS cohort_id", sql, fixed = TRUE)) {
      return(data.frame(
        cohortId = as.integer(cohort_ids),
        cohortEntries = rep(10L, length(cohort_ids)),
        cohortSubjects = rep(9L, length(cohort_ids))
      ))
    }

    if (grepl("SELECT * FROM #pa_stats", sql, fixed = TRUE)) {
      return(data.frame(pathwayAnalysisGenerationId = 1L, targetCohortId = 10L, countValue = 5L))
    }

    if (grepl("SELECT * FROM #pa_paths", sql, fixed = TRUE)) {
      return(data.frame(step1 = 0L, step2 = NA_integer_, countValue = 5L))
    }

    if (grepl("SELECT * FROM #pa_events", sql, fixed = TRUE)) {
      return(data.frame(eventCohortId = integer(), countValue = integer()))
    }

    stop("Unexpected SQL query in mock.")
  }

  get_tables_mock <- function(...) c("cohort")
  exec_mock <- function(...) invisible(NULL)

  result <- mock_dbconnector(
    mocks = list(
      getTableNames = get_tables_mock,
      renderTranslateQuerySql = query_mock,
      renderTranslateExecuteSql = exec_mock
    ),
    code = OdysseusPathwayModule::executeCohortPathways(
      connection = structure(list(), class = "mock_connection"),
      cohortDatabaseSchema = "main",
      cohortTableName = "cohort",
      targetCohortIds = 10L,
      eventCohortIds = c(1L, 2L),
      analysisType = "pre-index",
      lookbackStartDay = -365,
      lookbackEndDay = -1,
      maxDepth = 2,
      collapseWindow = 30
    )
  )

  expect_equal(nrow(result$pathwaycomboIds), 0)
  expect_equal(nrow(result$pathwayAnalysisCodesLong), 0)
  expect_equal(nrow(result$pathwayAnalysisCodesData), 0)
  expect_equal(nrow(result$isCombo), 0)
})

test_that("executeCohortPathways validates schema type before connecting", {
  expect_error(
    OdysseusPathwayModule::executeCohortPathways(
      cohortDatabaseSchema = 1,
      targetCohortIds = 1,
      eventCohortIds = 1
    ),
    "cohortDatabaseSchema"
  )
})

test_that("executeCohortPathways validates pre-index lookback window order", {
  expect_error(
    OdysseusPathwayModule::executeCohortPathways(
      cohortDatabaseSchema = "main",
      targetCohortIds = 1,
      eventCohortIds = 1,
      analysisType = "pre-index",
      lookbackStartDay = -1,
      lookbackEndDay = -365
    ),
    "must be less than"
  )
})

test_that("executeCohortPathways validates analysisType choices", {
  expect_error(
    OdysseusPathwayModule::executeCohortPathways(
      cohortDatabaseSchema = "main",
      targetCohortIds = 1,
      eventCohortIds = 1,
      analysisType = "during-index"
    ),
    "should be one of"
  )
})

test_that("executeCohortPathways validates numeric controls", {
  expect_error(
    OdysseusPathwayModule::executeCohortPathways(
      cohortDatabaseSchema = "main",
      targetCohortIds = 1,
      eventCohortIds = 1,
      minCellCount = -1
    ),
    "minCellCount"
  )

  expect_error(
    OdysseusPathwayModule::executeCohortPathways(
      cohortDatabaseSchema = "main",
      targetCohortIds = 1,
      eventCohortIds = 1,
      maxDepth = 0
    ),
    "maxDepth"
  )

  expect_error(
    OdysseusPathwayModule::executeCohortPathways(
      cohortDatabaseSchema = "main",
      targetCohortIds = 1,
      eventCohortIds = 1,
      collapseWindow = -10
    ),
    "collapseWindow"
  )
})

test_that("executeCohortPathways validates allowRepeats flag", {
  expect_error(
    OdysseusPathwayModule::executeCohortPathways(
      cohortDatabaseSchema = "main",
      targetCohortIds = 1,
      eventCohortIds = 1,
      allowRepeats = "FALSE"
    ),
    "allowRepeats"
  )
})

test_that("executeCohortPathways validates pre-index day sign", {
  expect_error(
    OdysseusPathwayModule::executeCohortPathways(
      cohortDatabaseSchema = "main",
      targetCohortIds = 1,
      eventCohortIds = 1,
      analysisType = "pre-index",
      lookbackStartDay = 0,
      lookbackEndDay = -1
    ),
    "lookbackStartDay"
  )

  expect_error(
    OdysseusPathwayModule::executeCohortPathways(
      cohortDatabaseSchema = "main",
      targetCohortIds = 1,
      eventCohortIds = 1,
      analysisType = "pre-index",
      lookbackStartDay = -30,
      lookbackEndDay = 0
    ),
    "lookbackEndDay"
  )
})

test_that("core package no longer exports plotting helpers", {
  exports <- getNamespaceExports("OdysseusPathwayModule")

  expect_false("createPathwaySunburst" %in% exports)
  expect_false("createPathwaySankey" %in% exports)
  expect_true("executeCohortPathways" %in% exports)
})

# Copyright 2026 Odysseus Data Services
#
# This file is part of OdysseusPathwayModule
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


#' Build an event sequence graph from pathway analysis results.
#'
#' @description
#' Takes the output of \code{\link{executeCohortPathways}} and a cohort name
#' mapping, decodes the bitmask-encoded combo IDs into human-readable event
#' names, and builds a directed \pkg{igraph} graph representing event
#' transitions across pathway steps.
#'
#' Nodes in the graph represent unique \emph{(event, step)} combinations.
#' Edges represent directed transitions between consecutive steps, weighted by
#' patient counts and annotated with transition probabilities.
#'
#' The returned object includes:
#' \itemize{
#'   \item \strong{graph}: A directed, weighted \code{\link[igraph]{igraph}}
#'         object. Vertex attributes: \code{name}, \code{eventName},
#'         \code{step}, \code{count}, \code{share}. Edge attributes:
#'         \code{weight} (= patient count), \code{probability},
#'         \code{sourceStep}, \code{targetStep}.
#'   \item \strong{sequences}: The fully decoded pathway sequences with a
#'         human-readable \code{pathway} string column.
#'   \item \strong{summary}: High-level statistics (total pathways, patients,
#'         depth, unique events, vertex/edge counts).
#'   \item \strong{eventNameLookup}: Named character vector mapping combo IDs
#'         to event names.
#' }
#'
#' @param cpResults   A list returned by \code{\link{executeCohortPathways}}.
#'                    Must contain at least \code{pathwaysAnalysisPathsData},
#'                    \code{pathwayAnalysisCodesLong}, and \code{isCombo}.
#' @param generationSet A data frame with at least two columns:
#'                      \code{cohortId} (integer) and \code{cohortName} (character).
#'                      Used to map event cohort IDs to descriptive names.
#' @param maxSteps    (Optional) Maximum number of steps to include. If \code{NULL}
#'                    (default), all non-empty steps are used.
#' @param minCount    (Default = 1) Minimum patient count for a pathway to be
#'                    included in the output.
#'
#' @return A list of class \code{"event_sequence_graph"} with components
#'         \code{graph} (igraph object), \code{sequences}, \code{summary}, and
#'         \code{eventNameLookup}.
#'
#' @examples
#' if (requireNamespace("igraph", quietly = TRUE)) {
#'   cpResults <- list(
#'     pathwaysAnalysisPathsData = data.frame(
#'       pathwayAnalysisGenerationId = c(1L, 1L),
#'       targetCohortId = c(10L, 10L),
#'       step1 = c(2L, 4L),
#'       step2 = c(4L, NA_integer_),
#'       countValue = c(10L, 5L)
#'     ),
#'     pathwayAnalysisCodesLong = data.frame(
#'       pathwayAnalysisGenerationId = c(1L, 1L),
#'       code = c(2L, 4L),
#'       targetCohortId = c(10L, 10L),
#'       eventCohortId = c(1L, 2L),
#'       isCombo = c(0L, 0L),
#'       numberOfEvents = c(1L, 1L)
#'     ),
#'     isCombo = data.frame(
#'       targetCohortId = c(10L, 10L),
#'       comboId = c(2L, 4L),
#'       numberOfEvents = c(1L, 1L),
#'       isCombo = c(0L, 0L)
#'     )
#'   )
#'
#'   generationSet <- data.frame(
#'     cohortId = c(1L, 2L, 10L),
#'     cohortName = c("Celecoxib", "Diclofenac", "NSAIDs")
#'   )
#'
#'   esg <- buildEventSequenceGraph(cpResults, generationSet)
#'   esg$summary
#'   igraph::gorder(esg$graph)
#' }
#'
#' @export
buildEventSequenceGraph <- function(cpResults,
                                    generationSet,
                                    maxSteps = NULL,
                                    minCount = 1) {

  check_installed("igraph", reason = "buildEventSequenceGraph() uses igraph for graph construction.")

  # --- Input validation ---
  if (!is.list(cpResults)) {
    stop("'cpResults' must be a list (output of executeCohortPathways).")
  }
  required <- c("pathwaysAnalysisPathsData", "pathwayAnalysisCodesLong", "isCombo")
  missing <- setdiff(required, names(cpResults))
  if (length(missing) > 0) {
    stop(
      "'cpResults' is missing required components: ",
      paste(missing, collapse = ", "), "."
    )
  }

  if (!is.data.frame(generationSet)) {
    stop("'generationSet' must be a data.frame.")
  }
  if (!all(c("cohortId", "cohortName") %in% names(generationSet))) {
    stop("'generationSet' must contain 'cohortId' and 'cohortName' columns.")
  }

  if (!is.null(maxSteps)) {
    if (!is.numeric(maxSteps) || length(maxSteps) != 1 || maxSteps < 1) {
      stop("'maxSteps' must be a single positive integer or NULL.")
    }
    maxSteps <- as.integer(maxSteps)
  }
  if (!is.numeric(minCount) || length(minCount) != 1 || minCount < 0) {
    stop("'minCount' must be a single non-negative number.")
  }

  # --- Build event name lookup: comboId -> human-readable name ---
  eventNameLookup <- .buildEventNameLookup(
    cpResults$pathwayAnalysisCodesLong,
    generationSet
  )

  # --- Extract and prepare paths data ---
  pathsData <- cpResults$pathwaysAnalysisPathsData
  stepCols <- grep("^step", names(pathsData), value = TRUE)

  # Remove all-NA step columns
  nonEmpty <- stepCols[colSums(!is.na(pathsData[, stepCols, drop = FALSE])) > 0]

  if (length(nonEmpty) == 0) {
    stop("No non-empty step columns found in pathwaysAnalysisPathsData.")
  }

  if (!is.null(maxSteps)) {
    nonEmpty <- nonEmpty[seq_len(min(maxSteps, length(nonEmpty)))]
  }

  # --- Decode step columns to event names ---
  stepNameCols <- character(length(nonEmpty))
  for (j in seq_along(nonEmpty)) {
    col <- nonEmpty[j]
    nameCol <- sub("^step", "stepName", col)
    stepNameCols[j] <- nameCol
    pathsData[[nameCol]] <- eventNameLookup[as.character(pathsData[[col]])]
  }

  # --- Filter by minCount ---
  pathsData <- pathsData[pathsData$countValue >= minCount, , drop = FALSE]
  if (nrow(pathsData) == 0) {
    stop("No pathways remain after filtering by minCount = ", minCount, ".")
  }

  # --- Build edges data frame (transitions between consecutive steps) ---
  edgesList <- list()
  for (i in seq_along(stepNameCols)[-1]) {
    prevCol <- stepNameCols[i - 1]
    currCol <- stepNameCols[i]

    rows <- !is.na(pathsData[[prevCol]]) & !is.na(pathsData[[currCol]])
    if (!any(rows)) next

    subDf <- data.frame(
      sourceName = pathsData[[prevCol]][rows],
      targetName = pathsData[[currCol]][rows],
      countValue = pathsData$countValue[rows],
      stringsAsFactors = FALSE
    )

    agg <- aggregate(countValue ~ sourceName + targetName, data = subDf, FUN = sum)
    agg$sourceStep <- i - 1L
    agg$targetStep <- as.integer(i)

    edgesList[[length(edgesList) + 1]] <- agg
  }

  if (length(edgesList) == 0) {
    stop("No transitions could be created. The data may only contain single-step pathways.")
  }

  edgesDf <- do.call(rbind, edgesList)
  rownames(edgesDf) <- NULL

  # Compute transition probabilities per source within each step transition
  edgesDf$probability <- NA_real_
  for (s in unique(edgesDf$sourceStep)) {
    idx <- edgesDf$sourceStep == s
    for (src in unique(edgesDf$sourceName[idx])) {
      srcIdx <- idx & edgesDf$sourceName == src
      total <- sum(edgesDf$countValue[srcIdx])
      edgesDf$probability[srcIdx] <- edgesDf$countValue[srcIdx] / total
    }
  }

  # --- Build nodes data frame ---
  nodesList <- list()
  for (i in seq_along(stepNameCols)) {
    col <- stepNameCols[i]
    notNA <- !is.na(pathsData[[col]])
    if (!any(notNA)) next

    tmp <- aggregate(
      pathsData$countValue[notNA],
      by = list(eventName = pathsData[[col]][notNA]),
      FUN = sum
    )
    names(tmp) <- c("eventName", "count")
    tmp$step <- as.integer(i)
    nodesList[[length(nodesList) + 1]] <- tmp
  }

  nodesDf <- do.call(rbind, nodesList)
  rownames(nodesDf) <- NULL

  # Compute share within each step
  nodesDf$share <- NA_real_
  for (s in unique(nodesDf$step)) {
    idx <- nodesDf$step == s
    total <- sum(nodesDf$count[idx])
    nodesDf$share[idx] <- nodesDf$count[idx] / total
  }

  # Create stable vertex IDs: "EventName [Step N]"
  nodesDf$name <- paste0(nodesDf$eventName, " [Step ", nodesDf$step, "]")

  # Map edges to vertex IDs
  edgesDf$from <- paste0(edgesDf$sourceName, " [Step ", edgesDf$sourceStep, "]")
  edgesDf$to   <- paste0(edgesDf$targetName, " [Step ", edgesDf$targetStep, "]")

  # --- Construct igraph object ---
  ig <- igraph::graph_from_data_frame(
    d        = edgesDf[, c("from", "to"), drop = FALSE],
    directed = TRUE,
    vertices = nodesDf[, c("name", "eventName", "step", "count", "share"), drop = FALSE]
  )

  # Set edge attributes
  igraph::E(ig)$weight      <- edgesDf$countValue
  igraph::E(ig)$probability <- edgesDf$probability
  igraph::E(ig)$sourceStep  <- edgesDf$sourceStep
  igraph::E(ig)$targetStep  <- edgesDf$targetStep

  # --- Build decoded sequences ---
  seqCols <- c(stepNameCols, "countValue")
  sequences <- pathsData[, seqCols, drop = FALSE]
  sequences$pathway <- apply(
    sequences[, stepNameCols, drop = FALSE], 1,
    function(row) paste(row[!is.na(row)], collapse = " -> ")
  )
  sequences$depth <- rowSums(!is.na(sequences[, stepNameCols, drop = FALSE]))
  rownames(sequences) <- NULL

  # --- Summary ---
  summaryInfo <- list(
    totalPathways = nrow(sequences),
    totalPatients = sum(sequences$countValue),
    maxDepth      = length(nonEmpty),
    uniqueEvents  = length(unique(nodesDf$eventName)),
    stepsUsed     = length(nonEmpty),
    vertexCount   = igraph::vcount(ig),
    edgeCount     = igraph::ecount(ig)
  )

  result <- list(
    graph           = ig,
    sequences       = sequences,
    summary         = summaryInfo,
    eventNameLookup = eventNameLookup
  )
  class(result) <- "event_sequence_graph"
  return(result)
}


#' @export
print.event_sequence_graph <- function(x, ...) {
  cat("Event Sequence Graph (igraph)\n")
  cat("-----------------------------\n")
  cat("Vertices (event-steps):", x$summary$vertexCount, "\n")
  cat("Edges (transitions):   ", x$summary$edgeCount, "\n")
  cat("Unique events:         ", x$summary$uniqueEvents, "\n")
  cat("Pathway depth:         ", x$summary$maxDepth, "steps\n")
  cat("Total pathways:        ", x$summary$totalPathways, "\n")
  cat("Total patients:        ", x$summary$totalPatients, "\n")
  cat("\nTop sequences:\n")
  topN <- min(10, nrow(x$sequences))
  ordered <- x$sequences[order(-x$sequences$countValue), ]
  for (i in seq_len(topN)) {
    cat(
      sprintf("  %d. [n=%d] %s\n", i, ordered$countValue[i], ordered$pathway[i])
    )
  }
  invisible(x)
}


#' Plot an event sequence graph.
#'
#' @description
#' Produces a layered plot of the event sequence graph using \code{\link[igraph]{plot.igraph}}.
#' Nodes are laid out by step (left to right), with edge widths proportional to
#' patient counts and node sizes proportional to event counts within each step.
#'
#' @param x An \code{event_sequence_graph} object (returned by
#'          \code{\link{buildEventSequenceGraph}}).
#' @param colorPalette Character vector of hex colors for nodes. If \code{NULL}
#'        (default), a built-in palette is used. Colors are mapped by unique event
#'        name (same event = same color across steps).
#' @param edgeWidthRange Numeric vector of length 2. Min and max edge widths
#'        (default: \code{c(0.5, 8)}).
#' @param vertexSizeRange Numeric vector of length 2. Min and max vertex sizes
#'        (default: \code{c(8, 25)}).
#' @param vertexLabelCex Label size multiplier (default: 0.7).
#' @param main Plot title (default: "Event Sequence Graph").
#' @param ... Additional arguments passed to \code{\link[igraph]{plot.igraph}}.
#'
#' @return Invisibly returns the igraph object.
#'
#' @examples
#' if (requireNamespace("igraph", quietly = TRUE)) {
#'   cpResults <- list(
#'     pathwaysAnalysisPathsData = data.frame(
#'       pathwayAnalysisGenerationId = c(1L, 1L),
#'       targetCohortId = c(10L, 10L),
#'       step1 = c(2L, 4L),
#'       step2 = c(4L, NA_integer_),
#'       countValue = c(10L, 5L)
#'     ),
#'     pathwayAnalysisCodesLong = data.frame(
#'       pathwayAnalysisGenerationId = c(1L, 1L),
#'       code = c(2L, 4L),
#'       targetCohortId = c(10L, 10L),
#'       eventCohortId = c(1L, 2L),
#'       isCombo = c(0L, 0L),
#'       numberOfEvents = c(1L, 1L)
#'     ),
#'     isCombo = data.frame(
#'       targetCohortId = c(10L, 10L),
#'       comboId = c(2L, 4L),
#'       numberOfEvents = c(1L, 1L),
#'       isCombo = c(0L, 0L)
#'     )
#'   )
#'
#'   generationSet <- data.frame(
#'     cohortId = c(1L, 2L, 10L),
#'     cohortName = c("Celecoxib", "Diclofenac", "NSAIDs")
#'   )
#'
#'   esg <- buildEventSequenceGraph(cpResults, generationSet)
#'   plot(esg)
#' }
#'
#' @export
plot.event_sequence_graph <- function(x,
                                      colorPalette = NULL,
                                      edgeWidthRange = c(0.5, 8),
                                      vertexSizeRange = c(8, 25),
                                      vertexLabelCex = 0.7,
                                      main = "Event Sequence Graph",
                                      ...) {

  check_installed("igraph", reason = "plot.event_sequence_graph() uses igraph for plotting.")


  ig <- x$graph

  # --- Layout: Sugiyama layered layout based on step attribute ---
  steps <- igraph::V(ig)$step
  layout <- igraph::layout_with_sugiyama(ig, layers = steps)$layout

  # --- Node colors: same event name -> same color ---
  uniqueEvents <- unique(igraph::V(ig)$eventName)

  if (is.null(colorPalette)) {
    colorPalette <- grDevices::hcl.colors(
      max(length(uniqueEvents), 3),
      palette = "Set 2"
    )
  }

  colorMap <- stats::setNames(
    rep_len(colorPalette, length(uniqueEvents)),
    uniqueEvents
  )
  vertexColors <- unname(colorMap[igraph::V(ig)$eventName])

  # --- Scale edge widths ---
  weights <- igraph::E(ig)$weight
  if (length(unique(weights)) == 1) {
    edgeWidths <- rep(mean(edgeWidthRange), length(weights))
  } else {
    edgeWidths <- edgeWidthRange[1] +
      (weights - min(weights)) / (max(weights) - min(weights)) *
      (edgeWidthRange[2] - edgeWidthRange[1])
  }

  # --- Scale vertex sizes ---
  counts <- igraph::V(ig)$count
  if (length(unique(counts)) == 1) {
    vertexSizes <- rep(mean(vertexSizeRange), length(counts))
  } else {
    vertexSizes <- vertexSizeRange[1] +
      (counts - min(counts)) / (max(counts) - min(counts)) *
      (vertexSizeRange[2] - vertexSizeRange[1])
  }

  # --- Edge colors: semi-transparent version of source node color ---
  edgeList <- igraph::as_edgelist(ig)
  sourceNames <- edgeList[, 1]
  sourceVertexIdx <- match(sourceNames, igraph::V(ig)$name)
  edgeColors <- vapply(vertexColors[sourceVertexIdx], function(col) {
    rgbVals <- grDevices::col2rgb(col)
    grDevices::rgb(rgbVals[1], rgbVals[2], rgbVals[3],
                   alpha = 100, maxColorValue = 255)
  }, character(1), USE.NAMES = FALSE)

  # --- Vertex labels: eventName only (drop step suffix) ---
  vertexLabels <- igraph::V(ig)$eventName

  # --- Plot ---
  igraph::plot.igraph(
    ig,
    layout        = layout,
    vertex.color  = vertexColors,
    vertex.size   = vertexSizes,
    vertex.label  = vertexLabels,
    vertex.label.cex = vertexLabelCex,
    vertex.label.color = "black",
    edge.width    = edgeWidths,
    edge.color    = edgeColors,
    edge.arrow.size = 0.5,
    edge.curved   = 0.1,
    main          = main,
    ...
  )

  invisible(ig)
}


# --- Internal helper ---

#' Build a named lookup vector: character(comboId) -> "EventA + EventB"
#' @keywords internal
.buildEventNameLookup <- function(codesLong, generationSet) {
  if (nrow(codesLong) == 0) {
    return(character(0))
  }

  # Map eventCohortId -> cohortName
  nameMap <- stats::setNames(
    as.character(generationSet$cohortName),
    as.character(generationSet$cohortId)
  )

  # Subset to unique code-eventCohortId pairs
  codes <- unique(codesLong[, c("code", "eventCohortId"), drop = FALSE])
  codes$eventName <- nameMap[as.character(codes$eventCohortId)]

  # For any unmapped IDs, use a fallback
  codes$eventName[is.na(codes$eventName)] <-
    paste0("Event_", codes$eventCohortId[is.na(codes$eventName)])

  # Aggregate: paste event names for each code (sorted for deterministic output)
  agg <- aggregate(
    eventName ~ code,
    data = codes,
    FUN = function(x) paste(sort(x), collapse = " + ")
  )

  stats::setNames(agg$eventName, as.character(agg$code))
}

# --- Shared test fixture ---

.make_cpResults <- function(pathsData, codesLong, isCombo,
                            comboIds = NULL) {
  if (is.null(comboIds)) {
    stepCols <- grep("^step", names(pathsData), value = TRUE)
    vals <- unlist(pathsData[, stepCols], use.names = FALSE)
    comboIds <- sort(unique(vals[!is.na(vals) & vals > 0]))
  }
  list(
    pathwayAnalysisStatsData = data.frame(
      pathwayAnalysisGenerationId = 1L, targetCohortId = 10L, countValue = 20L
    ),
    pathwaysAnalysisPathsData = pathsData,
    pathwaysAnalysisEventsData = data.frame(eventCohortId = 1L, countValue = 10L),
    pathwaycomboIds = data.frame(comboIds = comboIds),
    pathwayAnalysisCodesLong = codesLong,
    isCombo = isCombo,
    pathwayAnalysisCodesData = data.frame(
      pathwayAnalysisGenerationId = rep(1L, nrow(codesLong)),
      code = codesLong$code,
      isCombo = codesLong$isCombo
    )
  )
}

.basic_generationSet <- data.frame(
  cohortId  = c(1L, 2L, 10L),
  cohortName = c("Celecoxib", "Diclofenac", "NSAIDs"),
  stringsAsFactors = FALSE
)

.basic_codesLong <- data.frame(
  pathwayAnalysisGenerationId = c(1L, 1L),
  code = c(2L, 4L),
  targetCohortId = c(10L, 10L),
  eventCohortId = c(1L, 2L),
  isCombo = c(0L, 0L),
  numberOfEvents = c(1L, 1L)
)

.basic_isCombo <- data.frame(
  targetCohortId = c(10L, 10L),
  comboId = c(2L, 4L),
  numberOfEvents = c(1L, 1L),
  isCombo = c(0L, 0L)
)


# --- Tests ---

test_that("buildEventSequenceGraph returns igraph-based structure", {
  cpResults <- .make_cpResults(
    pathsData = data.frame(
      pathwayAnalysisGenerationId = c(1L, 1L, 1L),
      targetCohortId = c(10L, 10L, 10L),
      step1 = c(2L, 4L, 2L),
      step2 = c(4L, 2L, NA),
      step3 = c(NA, NA, NA),
      countValue = c(10L, 5L, 3L)
    ),
    codesLong = .basic_codesLong,
    isCombo   = .basic_isCombo
  )

  esg <- buildEventSequenceGraph(cpResults, .basic_generationSet)

  expect_s3_class(esg, "event_sequence_graph")
  expect_named(esg, c("graph", "sequences", "summary", "eventNameLookup"))

  # igraph object
  expect_s3_class(esg$graph, "igraph")
  expect_true(igraph::is_directed(esg$graph))

  # Vertex attributes
  expect_true("eventName" %in% igraph::vertex_attr_names(esg$graph))
  expect_true("step"      %in% igraph::vertex_attr_names(esg$graph))
  expect_true("count"     %in% igraph::vertex_attr_names(esg$graph))
  expect_true("share"     %in% igraph::vertex_attr_names(esg$graph))

  # Edge attributes
  expect_true("weight"      %in% igraph::edge_attr_names(esg$graph))
  expect_true("probability" %in% igraph::edge_attr_names(esg$graph))
  expect_true("sourceStep"  %in% igraph::edge_attr_names(esg$graph))
  expect_true("targetStep"  %in% igraph::edge_attr_names(esg$graph))

  # Sequences
  expect_s3_class(esg$sequences, "data.frame")
  expect_true("pathway" %in% names(esg$sequences))
  expect_true("depth"   %in% names(esg$sequences))

  # Summary
  expect_equal(esg$summary$totalPathways, 3)
  expect_equal(esg$summary$totalPatients, 18)
  expect_equal(esg$summary$uniqueEvents, 2)
  expect_true(esg$summary$vertexCount > 0)
  expect_true(esg$summary$edgeCount > 0)
})


test_that("igraph vertex names encode event + step", {
  cpResults <- .make_cpResults(
    pathsData = data.frame(
      pathwayAnalysisGenerationId = 1L,
      targetCohortId = 10L,
      step1 = 2L, step2 = 4L,
      countValue = 10L
    ),
    codesLong = .basic_codesLong,
    isCombo   = .basic_isCombo
  )

  esg <- buildEventSequenceGraph(cpResults, .basic_generationSet)
  vnames <- igraph::V(esg$graph)$name

  expect_true(all(grepl("\\[Step \\d+\\]$", vnames)))
  expect_true("Celecoxib [Step 1]" %in% vnames)
  expect_true("Diclofenac [Step 2]" %in% vnames)
})


test_that("igraph edge weights match patient counts and probabilities sum to 1", {
  cpResults <- .make_cpResults(
    pathsData = data.frame(
      pathwayAnalysisGenerationId = c(1L, 1L, 1L),
      targetCohortId = c(10L, 10L, 10L),
      step1 = c(2L, 2L, 4L),
      step2 = c(4L, 2L, 2L),
      countValue = c(10L, 8L, 5L)
    ),
    codesLong = .basic_codesLong,
    isCombo   = .basic_isCombo
  )

  esg <- buildEventSequenceGraph(cpResults, .basic_generationSet)
  ig <- esg$graph

  # Weights are positive integers
  expect_true(all(igraph::E(ig)$weight > 0))

  # Probabilities per source-step must sum to 1
  el <- igraph::as_data_frame(ig, what = "edges")
  for (s in unique(el$sourceStep)) {
    stepEdges <- el[el$sourceStep == s, ]
    for (src in unique(igraph::tail_of(ig, igraph::E(ig)[el$sourceStep == s])$name)) {
      srcEdges <- el[el$sourceStep == s & el$from == src, ]
      expect_equal(sum(srcEdges$probability), 1, tolerance = 1e-10)
    }
  }
})


test_that("buildEventSequenceGraph decodes combo events correctly", {
  comboCodesLong <- data.frame(
    pathwayAnalysisGenerationId = c(1L, 1L, 1L, 1L),
    code = c(2L, 4L, 6L, 6L),
    targetCohortId = c(10L, 10L, 10L, 10L),
    eventCohortId = c(1L, 2L, 1L, 2L),
    isCombo = c(0L, 0L, 1L, 1L),
    numberOfEvents = c(1L, 1L, 2L, 2L)
  )
  comboIsCombo <- data.frame(
    targetCohortId = c(10L, 10L, 10L),
    comboId = c(2L, 4L, 6L),
    numberOfEvents = c(1L, 1L, 2L),
    isCombo = c(0L, 0L, 1L)
  )

  cpResults <- .make_cpResults(
    pathsData = data.frame(
      pathwayAnalysisGenerationId = 1L,
      targetCohortId = 10L,
      step1 = 6L, step2 = 2L,
      countValue = 10L
    ),
    codesLong = comboCodesLong,
    isCombo   = comboIsCombo
  )

  esg <- buildEventSequenceGraph(cpResults, .basic_generationSet)

  expect_equal(esg$eventNameLookup[["6"]], "Celecoxib + Diclofenac")
  expect_equal(esg$eventNameLookup[["2"]], "Celecoxib")
  expect_true(any(grepl("Celecoxib \\+ Diclofenac", esg$sequences$pathway)))
  expect_true(any(grepl("Celecoxib \\+ Diclofenac", igraph::V(esg$graph)$eventName)))
})


test_that("buildEventSequenceGraph respects maxSteps parameter", {
  cpResults <- .make_cpResults(
    pathsData = data.frame(
      pathwayAnalysisGenerationId = 1L,
      targetCohortId = 10L,
      step1 = 2L, step2 = 4L, step3 = 2L,
      countValue = 10L
    ),
    codesLong = .basic_codesLong,
    isCombo   = .basic_isCombo
  )

  esg <- buildEventSequenceGraph(cpResults, .basic_generationSet, maxSteps = 2)

  expect_equal(esg$summary$maxDepth, 2)
  expect_true(all(igraph::E(esg$graph)$sourceStep == 1))
  expect_true(all(igraph::E(esg$graph)$targetStep == 2))
  expect_true(max(igraph::V(esg$graph)$step) <= 2)
})


test_that("buildEventSequenceGraph filters by minCount", {
  cpResults <- .make_cpResults(
    pathsData = data.frame(
      pathwayAnalysisGenerationId = c(1L, 1L),
      targetCohortId = c(10L, 10L),
      step1 = c(2L, 4L),
      step2 = c(4L, 2L),
      countValue = c(15L, 3L)
    ),
    codesLong = .basic_codesLong,
    isCombo   = .basic_isCombo
  )

  esg <- buildEventSequenceGraph(cpResults, .basic_generationSet, minCount = 5)

  expect_equal(esg$summary$totalPathways, 1)
  expect_equal(esg$summary$totalPatients, 15)
})


test_that("igraph supports standard downstream analysis", {
  cpResults <- .make_cpResults(
    pathsData = data.frame(
      pathwayAnalysisGenerationId = c(1L, 1L),
      targetCohortId = c(10L, 10L),
      step1 = c(2L, 4L),
      step2 = c(4L, 2L),
      countValue = c(10L, 5L)
    ),
    codesLong = .basic_codesLong,
    isCombo   = .basic_isCombo
  )

  esg <- buildEventSequenceGraph(cpResults, .basic_generationSet)
  ig <- esg$graph

  # as_data_frame round-trip
  edgeDf <- igraph::as_data_frame(ig, what = "edges")
  expect_true("weight" %in% names(edgeDf))
  expect_true("probability" %in% names(edgeDf))

  vertDf <- igraph::as_data_frame(ig, what = "vertices")
  expect_true("eventName" %in% names(vertDf))
  expect_true("step" %in% names(vertDf))

  # degree
  outDeg <- igraph::degree(ig, mode = "out")
  inDeg  <- igraph::degree(ig, mode = "in")
  expect_type(outDeg, "double")
  expect_type(inDeg, "double")

  # shortest paths (inverted weights)
  dists <- igraph::distances(ig, weights = 1 / igraph::E(ig)$weight)
  expect_true(is.matrix(dists))

  # neighborhood
  nbrs <- igraph::neighbors(ig, igraph::V(ig)[1], mode = "out")
  expect_true(length(nbrs) >= 0)
})


test_that("igraph vertex shares sum to 1 within each step", {
  cpResults <- .make_cpResults(
    pathsData = data.frame(
      pathwayAnalysisGenerationId = c(1L, 1L, 1L),
      targetCohortId = c(10L, 10L, 10L),
      step1 = c(2L, 2L, 4L),
      step2 = c(4L, 2L, 2L),
      countValue = c(10L, 8L, 5L)
    ),
    codesLong = .basic_codesLong,
    isCombo   = .basic_isCombo
  )

  esg <- buildEventSequenceGraph(cpResults, .basic_generationSet)
  vdf <- igraph::as_data_frame(esg$graph, what = "vertices")

  for (s in unique(vdf$step)) {
    stepVerts <- vdf[vdf$step == s, ]
    expect_equal(sum(stepVerts$share), 1, tolerance = 1e-10)
  }
})


test_that("buildEventSequenceGraph errors on missing cpResults components", {
  expect_error(
    buildEventSequenceGraph(
      cpResults = list(pathwaysAnalysisPathsData = data.frame()),
      generationSet = .basic_generationSet
    ),
    "missing required components"
  )
})


test_that("buildEventSequenceGraph errors on invalid generationSet", {
  cpResults <- list(
    pathwaysAnalysisPathsData = data.frame(),
    pathwayAnalysisCodesLong = data.frame(),
    isCombo = data.frame()
  )

  expect_error(
    buildEventSequenceGraph(cpResults, generationSet = data.frame(id = 1, name = "A")),
    "cohortId.*cohortName"
  )

  expect_error(
    buildEventSequenceGraph(cpResults, generationSet = "not a data frame"),
    "data.frame"
  )
})


test_that("print.event_sequence_graph produces output", {
  cpResults <- .make_cpResults(
    pathsData = data.frame(
      pathwayAnalysisGenerationId = 1L,
      targetCohortId = 10L,
      step1 = 2L, step2 = 4L,
      countValue = 10L
    ),
    codesLong = .basic_codesLong,
    isCombo   = .basic_isCombo
  )

  esg <- buildEventSequenceGraph(cpResults, .basic_generationSet)

  output <- capture.output(print(esg))
  expect_true(any(grepl("Event Sequence Graph", output)))
  expect_true(any(grepl("igraph", output)))
  expect_true(any(grepl("Celecoxib", output)))
  expect_true(any(grepl("Vertices", output)))
  expect_true(any(grepl("Edges", output)))
})


test_that("plot.event_sequence_graph runs without error", {
  cpResults <- .make_cpResults(
    pathsData = data.frame(
      pathwayAnalysisGenerationId = c(1L, 1L),
      targetCohortId = c(10L, 10L),
      step1 = c(2L, 4L),
      step2 = c(4L, 2L),
      countValue = c(10L, 5L)
    ),
    codesLong = .basic_codesLong,
    isCombo   = .basic_isCombo
  )

  esg <- buildEventSequenceGraph(cpResults, .basic_generationSet)

  # Plot to null device
  grDevices::pdf(NULL)
  on.exit(grDevices::dev.off(), add = TRUE)

  result <- plot(esg)

  # Returns the igraph invisibly
  expect_s3_class(result, "igraph")
})


test_that("buildEventSequenceGraph errors when cpResults is not a list", {
  expect_error(
    buildEventSequenceGraph(cpResults = "not_a_list", generationSet = .basic_generationSet),
    "must be a list"
  )
})

test_that("buildEventSequenceGraph errors on invalid maxSteps", {
  cpResults <- .make_cpResults(
    pathsData = data.frame(
      pathwayAnalysisGenerationId = 1L, targetCohortId = 10L,
      step1 = 2L, step2 = 4L, countValue = 10L
    ),
    codesLong = .basic_codesLong, isCombo = .basic_isCombo
  )
  expect_error(
    buildEventSequenceGraph(cpResults, .basic_generationSet, maxSteps = -1),
    "maxSteps"
  )
})

test_that("buildEventSequenceGraph errors on invalid minCount", {
  cpResults <- .make_cpResults(
    pathsData = data.frame(
      pathwayAnalysisGenerationId = 1L, targetCohortId = 10L,
      step1 = 2L, step2 = 4L, countValue = 10L
    ),
    codesLong = .basic_codesLong, isCombo = .basic_isCombo
  )
  expect_error(
    buildEventSequenceGraph(cpResults, .basic_generationSet, minCount = -5),
    "minCount"
  )
})

test_that("buildEventSequenceGraph errors when all step columns are NA", {
  cpResults <- .make_cpResults(
    pathsData = data.frame(
      pathwayAnalysisGenerationId = 1L, targetCohortId = 10L,
      step1 = NA_integer_, step2 = NA_integer_, countValue = 10L
    ),
    codesLong = .basic_codesLong, isCombo = .basic_isCombo,
    comboIds = 2L
  )
  expect_error(
    buildEventSequenceGraph(cpResults, .basic_generationSet),
    "No non-empty step columns"
  )
})

test_that("buildEventSequenceGraph errors when minCount filters all rows", {
  cpResults <- .make_cpResults(
    pathsData = data.frame(
      pathwayAnalysisGenerationId = 1L, targetCohortId = 10L,
      step1 = 2L, step2 = 4L, countValue = 1L
    ),
    codesLong = .basic_codesLong, isCombo = .basic_isCombo
  )
  expect_error(
    buildEventSequenceGraph(cpResults, .basic_generationSet, minCount = 100),
    "No pathways remain"
  )
})

test_that("buildEventSequenceGraph errors for single-step-only data", {
  cpResults <- .make_cpResults(
    pathsData = data.frame(
      pathwayAnalysisGenerationId = 1L, targetCohortId = 10L,
      step1 = 2L, countValue = 10L
    ),
    codesLong = .basic_codesLong, isCombo = .basic_isCombo
  )
  expect_error(
    buildEventSequenceGraph(cpResults, .basic_generationSet),
    "No transitions could be created"
  )
})

test_that("plot.event_sequence_graph handles uniform weights and counts", {
  cpResults <- .make_cpResults(
    pathsData = data.frame(
      pathwayAnalysisGenerationId = c(1L, 1L),
      targetCohortId = c(10L, 10L),
      step1 = c(2L, 2L),
      step2 = c(4L, 4L),
      countValue = c(10L, 10L)
    ),
    codesLong = .basic_codesLong, isCombo = .basic_isCombo
  )

  esg <- buildEventSequenceGraph(cpResults, .basic_generationSet)

  grDevices::pdf(NULL)
  on.exit(grDevices::dev.off(), add = TRUE)

  result <- plot(esg, main = "Uniform test")
  expect_s3_class(result, "igraph")
})

# Copyright 2026 Odysseus Data Services
#
# This file is part of OdysseusPathwayModule
#
# Adapted from CohortPathways by OHDSI (https://github.com/OHDSI/CohortPathways).
# The core implementation of CohortPathway analysis was developed by Christopher
# Knoll (github: chrisknoll), and enhanced by members of the OHDSI community:
# mick-iqvia (github), Odysseus Inc., and many other collaborators on the
# OHDSI/WebAPI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# --- Standalone helpers (replaces purrr dependency) ---
# Adapted from upstream-src/standalone.R



#' Safely Extract Nested Elements
#'
#' A lightweight replacement for \code{purrr::pluck()} that extracts
#' deeply nested elements from lists using integer positions, character
#' names, or accessor functions.
#'
#' @param .x A list or vector to extract from.
#' @param ... Accessors (integer, character, or function) applied sequentially.
#' @param .default Value to return if the element is missing or \code{NULL}.
#'
#' @return The extracted element, or \code{.default} if not found.
#'
#' @keywords internal
#' @noRd
.pluck <- function(.x, ..., .default = NULL) {
  accessors <- list(...)
  if (length(accessors) == 0) return(.x)

  result <- .x
  for (accessor in accessors) {
    if (is.null(result)) return(.default)
    if (is.numeric(accessor)) {
      accessor <- as.integer(accessor)
      if (accessor > length(result) || accessor < 1) return(.default)
      result <- result[[accessor]]
    } else if (is.character(accessor)) {
      if (!accessor %in% names(result)) return(.default)
      result <- result[[accessor]]
    } else if (is.function(accessor)) {
      result <- accessor(result)
    } else {
      stop("Accessor must be an integer, character, or function")
    }
  }
  if (is.null(result)) return(.default)
  result
}

#' Map and Column-Bind Results
#'
#' A lightweight replacement for \code{purrr::map_dfc()} that applies a
#' function to each element of a list and column-binds the results into a
#' single data frame.
#'
#' @param .x A list or vector to iterate over.
#' @param .f A function to apply to each element of \code{.x}.
#' @param ... Additional arguments passed to \code{.f}.
#'
#' @return A \code{data.frame} formed by column-binding the results.
#'
#' @keywords internal
#' @noRd
.map_dfc <- function(.x, .f, ...) {
  result_list <- lapply(.x, .f, ...)
  do.call(cbind, lapply(result_list, as.data.frame))
}


# --- Package check utility ---

#' Check Required Packages and Versions
#'
#' Verifies that the specified packages are installed and optionally meet
#' minimum version requirements. In interactive sessions, offers to install
#' missing packages; in non-interactive sessions, throws an informative error.
#'
#' @param pkg A character vector of package names to check for installation
#'   and version requirements.
#' @param reason Optional string providing context for why the packages are
#'   needed, included in error messages.
#' @param version Optional character vector of required package versions
#'   corresponding to \code{pkg}. Can be a single version or one per package.
#' @param compare Optional character vector of comparison operators (e.g.
#'   \code{">="}, \code{"=="}) corresponding to \code{pkg} for version checks.
#'   Can be a single operator or one per package. Defaults to \code{">="}
#'   if \code{version} is provided without \code{compare}.
#' @param action Optional function to call if any packages are missing or do
#'   not meet version requirements. The function will be passed the arguments
#'   \code{pkg}, \code{version}, \code{compare}, and \code{reason} for custom
#'   handling (e.g., automated installation). If not provided, an interactive
#'   prompt will be offered to install missing packages, or an error will be
#'   thrown in non-interactive sessions.
#' @param call The call to report in error messages if packages are missing.
#'   Defaults to the parent call.
#'
#' @return Invisibly returns \code{TRUE} if all packages are installed and
#'   meet version requirements. Otherwise, throws an error or invokes
#'   \code{action}.
#'
#' @keywords internal
#' @noRd
check_installed <- function(pkg,
                            reason = NULL,
                            version = NULL,
                            compare = NULL,
                            action = NULL,
                            call = sys.call(-1L)) {


  # --- input validation ---

  if (!is.character(pkg) || length(pkg) == 0L) {
    stop("`pkg` must be a character vector of package names.", call. = FALSE)
  }

  if (!is.null(version)) {
    version <- rep_len(as.character(version), length(pkg))
  }

  if (!is.null(compare)) {
    compare <- rep_len(as.character(compare), length(pkg))
  } else if (!is.null(version)) {
    compare <- rep_len(">=", length(pkg))
  }

  # --- check each package ---
  missing_pkg     <- character(0L)
  needs_update    <- character(0L)
  missing_reasons <- character(0L)

  for (i in seq_along(pkg)) {
    p <- pkg[i]

    if (!requireNamespace(p, quietly = TRUE)) {
      missing_pkg <- c(missing_pkg, p)
      next
    }

    # version check
    if (!is.null(version) && !is.na(version[i])) {
      cmp <- if (!is.null(compare)) compare[i] else ">="
      current <- as.character(utils::packageVersion(p))
      satisfied <- utils::compareVersion(current, version[i])

      ok <- switch(cmp,
                   ">="  = satisfied >= 0L,
                   ">"   = satisfied >  0L,
                   "=="  = satisfied == 0L,
                   "<="  = satisfied <= 0L,
                   "<"   = satisfied <  0L,
                   "!="  = satisfied != 0L,
                   stop(sprintf("Unknown comparison operator `%s`.", cmp),
                        call. = FALSE)
      )

      if (!ok) {
        needs_update <- c(needs_update, p)
        missing_reasons <- c(
          missing_reasons,
          sprintf(
            "  - %s (installed: %s, required: %s %s)",
            p, current, cmp, version[i]
          )
        )
      }
    }
  }

  # --- nothing to do ---
  if (length(missing_pkg) == 0L && length(needs_update) == 0L) {
    return(invisible(TRUE))
  }

  # --- build informative message ---
  parts <- character(0L)

  if (length(missing_pkg) > 0L) {
    noun <- if (length(missing_pkg) == 1L) "package" else "packages"
    parts <- c(parts, sprintf(
      "The following %s %s required but not installed:\n  - %s",
      noun,
      if (length(missing_pkg) == 1L) "is" else "are",
      paste(missing_pkg, collapse = "\n  - ")
    ))
  }

  if (length(needs_update) > 0L) {
    noun <- if (length(needs_update) == 1L) "package" else "packages"
    parts <- c(parts, sprintf(
      "The following %s %s installed but %s an update:\n%s",
      noun,
      if (length(needs_update) == 1L) "is" else "are",
      if (length(needs_update) == 1L) "needs" else "need",
      paste(missing_reasons, collapse = "\n")
    ))
  }

  msg <- paste(parts, collapse = "\n\n")

  if (!is.null(reason)) {
    msg <- paste0(msg, "\n\nReason: ", reason)
  }

  # --- custom action ---
  if (is.function(action)) {
    action(pkg      = c(missing_pkg, needs_update),
           version  = version,
           compare  = compare,
           reason   = reason)
    return(invisible(FALSE))
  }

  # --- interactive: offer to install ---
  all_needed <- unique(c(missing_pkg, needs_update))

  if (interactive()) {
    message(msg)
    ans <- utils::menu(
      choices = c("Yes", "No"),
      title   = sprintf(
        "\nWould you like to install/update %s now?",
        paste(all_needed, collapse = ", ")
      )
    )

    if (identical(ans, 1L)) {
      utils::install.packages(all_needed)

      # re-check after installation
      still_bad <- character(0L)
      for (p in all_needed) {
        if (!requireNamespace(p, quietly = TRUE)) {
          still_bad <- c(still_bad, p)
        }
      }
      if (length(still_bad) > 0L) {
        stop(
          sprintf(
            "Installation failed for: %s",
            paste(still_bad, collapse = ", ")
          ),
          call. = FALSE
        )
      }
      return(invisible(TRUE))
    }
  }

  # --- non-interactive or user declined ---
  install_cmd <- sprintf(
    'install.packages(%s)',
    if (length(all_needed) == 1L) {
      sprintf('"%s"', all_needed)
    } else {
      sprintf('c(%s)', paste(sprintf('"%s"', all_needed), collapse = ", "))
    }
  )

  stop(
    sprintf("%s\n\nRun `%s` to install.", msg, install_cmd),
    call. = FALSE
  )
}


# --- Core utility ---

#' Extract Bit-Position Components from a Sum of Powers of Two
#'
#' Given an integer \code{x} that represents a sum of distinct powers of two,
#' this function returns the zero-based bit positions (exponents) that compose
#' the sum. For example, \code{extractBitSum(5)} returns \code{c(2, 0)}
#' because \eqn{5 = 2^2 + 2^0}.
#'
#' @param x A non-negative integer representing a sum of powers of two.
#'
#' @return An integer vector of zero-based bit positions (exponents) whose
#'   corresponding powers of two sum to \code{x}, returned in descending order.
#'
#' @keywords internal
#' @noRd
extractBitSum <- function(x) {
  lengthVar <- round(log(x = x, base = 2) + 2)
  series <- c(2^(0:lengthVar))
  remainder <- x
  combination <- c()
  while (remainder != 0) {
    component <- match(TRUE, series > remainder) - 1
    remainder <- remainder - series[component]
    combination[length(combination) + 1] <- component - 1
  }
  return(combination)
}

# --- Tests for check_installed() ---

test_that("check_installed rejects non-character or empty pkg", {
  expect_error(
    OdysseusPathwayModule:::check_installed(pkg = 123),
    "`pkg` must be a character vector"
  )
  expect_error(
    OdysseusPathwayModule:::check_installed(pkg = character(0)),
    "`pkg` must be a character vector"
  )
})

test_that("check_installed returns TRUE for an installed package", {
  expect_true(OdysseusPathwayModule:::check_installed("testthat"))
})

test_that("check_installed returns TRUE when version is satisfied (>=)", {
  expect_true(
    OdysseusPathwayModule:::check_installed("testthat", version = "0.0.1", compare = ">=")
  )
})

test_that("check_installed defaults compare to >= when version supplied", {

  expect_true(
    OdysseusPathwayModule:::check_installed("testthat", version = "0.0.1")
  )
})

test_that("check_installed version check with == operator succeeds", {
  current <- as.character(utils::packageVersion("testthat"))
  expect_true(
    OdysseusPathwayModule:::check_installed("testthat", version = current, compare = "==")
  )
})

test_that("check_installed version check with <= operator succeeds", {
  expect_true(
    OdysseusPathwayModule:::check_installed("testthat", version = "99999.0.0", compare = "<=")
  )
})

test_that("check_installed version check with < operator succeeds", {
  expect_true(
    OdysseusPathwayModule:::check_installed("testthat", version = "99999.0.0", compare = "<")
  )
})

test_that("check_installed version check with > operator on low version", {
  expect_true(
    OdysseusPathwayModule:::check_installed("testthat", version = "0.0.1", compare = ">")
  )
})

test_that("check_installed version check with != operator", {
  expect_true(
    OdysseusPathwayModule:::check_installed("testthat", version = "0.0.1", compare = "!=")
  )
})

test_that("check_installed errors on unknown comparison operator", {
  expect_error(
    OdysseusPathwayModule:::check_installed("testthat", version = "0.0.1", compare = "~="),
    "Unknown comparison operator"
  )
})

test_that("check_installed stops for single missing package (non-interactive)", {
  err <- expect_error(
    OdysseusPathwayModule:::check_installed("thisNotExistPkg999xyz"),
    "required but not installed"
  )
  expect_match(conditionMessage(err), 'install.packages\\("thisNotExistPkg999xyz"\\)')
  expect_match(conditionMessage(err), "package is required")
})

test_that("check_installed stops for multiple missing packages (non-interactive)", {
  err <- expect_error(
    OdysseusPathwayModule:::check_installed(c("noSuchA999", "noSuchB999")),
    "packages are required"
  )
  expect_match(conditionMessage(err), "install.packages\\(c\\(")
})

test_that("check_installed includes reason in error message", {
  expect_error(
    OdysseusPathwayModule:::check_installed(
      "thisNotExistPkg999xyz",
      reason = "needed for testing"
    ),
    "Reason: needed for testing"
  )
})

test_that("check_installed stops when version needs update (single pkg)", {
  err <- expect_error(
    OdysseusPathwayModule:::check_installed("testthat", version = "99999.0.0", compare = ">="),
    "needs an update"
  )
  expect_match(conditionMessage(err), "installed:")
  expect_match(conditionMessage(err), "required: >= 99999.0.0")
})

test_that("check_installed stops when version needs update (multiple pkgs)", {
  # Use two packages guaranteed to be installed during tests
  err <- expect_error(
    OdysseusPathwayModule:::check_installed(
      c("testthat", "utils"),
      version = "99999.0.0",
      compare = ">="
    ),
    "need an update"
  )
  expect_match(conditionMessage(err), "packages")
})

test_that("check_installed invokes custom action and returns FALSE", {
  captured <- NULL
  result <- OdysseusPathwayModule:::check_installed(
    "thisNotExistPkg999xyz",
    action = function(pkg, version, compare, reason) {
      captured <<- list(pkg = pkg, version = version,
                        compare = compare, reason = reason)
    }
  )
  expect_false(result)
  expect_equal(captured$pkg, "thisNotExistPkg999xyz")
})

test_that("check_installed custom action receives reason", {
  captured_reason <- NULL
  OdysseusPathwayModule:::check_installed(
    "thisNotExistPkg999xyz",
    reason = "testing reason",
    action = function(pkg, version, compare, reason) {
      captured_reason <<- reason
    }
  )
  expect_equal(captured_reason, "testing reason")
})

test_that("check_installed custom action for version-update path", {
  captured_pkgs <- NULL
  OdysseusPathwayModule:::check_installed(
    "testthat",
    version = "99999.0.0",
    compare = ">=",
    action = function(pkg, version, compare, reason) {
      captured_pkgs <<- pkg
    }
  )
  expect_equal(captured_pkgs, "testthat")
})

test_that("check_installed with both missing and needs-update uses action", {
  captured_pkgs <- NULL
  OdysseusPathwayModule:::check_installed(
    c("thisNotExistPkg999xyz", "testthat"),
    version = c(NA, "99999.0.0"),
    compare = c(">=", ">="),
    action = function(pkg, version, compare, reason) {
      captured_pkgs <<- pkg
    }
  )
  expect_true("thisNotExistPkg999xyz" %in% captured_pkgs)
  expect_true("testthat" %in% captured_pkgs)
})

test_that("check_installed version NA is skipped", {
  expect_true(
    OdysseusPathwayModule:::check_installed("testthat", version = NA_character_)
  )
})

test_that("internal .pluck handles nested extraction and defaults", {
  x <- list(a = list(b = list(c = 42)))

  expect_equal(OdysseusPathwayModule:::.pluck(x, "a", "b", "c"), 42)
  expect_equal(OdysseusPathwayModule:::.pluck(x, "a", "missing", .default = "x"), "x")
})

test_that("internal .map_dfc combines outputs by column", {
  cols <- OdysseusPathwayModule:::.map_dfc(
    c("u", "v"),
    function(tag) data.frame(value = paste0(tag, 1:2), stringsAsFactors = FALSE)
  )

  expect_s3_class(cols, "data.frame")
  expect_equal(ncol(cols), 2)
  expect_equal(unname(cols[[1]]), c("u1", "u2"))
  expect_equal(unname(cols[[2]]), c("v1", "v2"))
})

test_that(".pluck with function accessor", {
  x <- list(a = 1:5)
  expect_equal(OdysseusPathwayModule:::.pluck(x, "a", length), 5)
})

test_that(".pluck returns .default for NULL element", {
  x <- list(a = NULL)
  expect_equal(OdysseusPathwayModule:::.pluck(x, "a", .default = "fb"), "fb")
})

test_that(".pluck returns .default for out-of-bounds numeric accessor", {
  expect_equal(OdysseusPathwayModule:::.pluck(list(1, 2), 10, .default = "oob"), "oob")
  expect_equal(OdysseusPathwayModule:::.pluck(list(1, 2), 0, .default = "oob"), "oob")
})

test_that(".pluck errors on invalid accessor type", {
  expect_error(
    OdysseusPathwayModule:::.pluck(list(a = 1), TRUE),
    "Accessor must be an integer, character, or function"
  )
})

test_that("extractBitSum decomposes powers-of-two combinations", {
  expect_equal(OdysseusPathwayModule:::extractBitSum(2), 1)
  expect_equal(OdysseusPathwayModule:::extractBitSum(6), c(2, 1))
  expect_equal(OdysseusPathwayModule:::extractBitSum(10), c(3, 1))
})

---
title: "Using OdysseusPathwayModule"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Using OdysseusPathwayModule}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Overview

OdysseusPathwayModule provides cohort pathway analysis for pre-instantiated
OMOP cohorts. The package focuses on one core workflow:

1. create or identify a target cohort and one or more event cohorts,
2. run pathway analysis with `executeCohortPathways()`, and
3. inspect pathway sequences, counts, and event-code mappings.

The package supports two analysis modes:

1. **Post-index** (`analysisType = "post-index"`, default): events occurring 
   *after* the target cohort index date.
2. **Pre-index** (`analysisType = "pre-index"`): events occurring *before* the 
   target cohort index date in a configurable lookback window.

Target and event cohorts can reside in the same cohort table or in separate
tables and schemas.

## Setup

```{r setup, eval=FALSE}
library(OdysseusPathwayModule)
library(Eunomia)

connectionDetails <- Eunomia::getEunomiaConnectionDetails()
```

## Create Example Cohorts with Eunomia

```{r cohorts, eval=FALSE}
generationSet <- Eunomia::createCohorts(connectionDetails)

generationSet
```

In the standard Eunomia example database, `createCohorts()` materializes four
cohorts in `main.cohort`:

- `1`: Celecoxib
- `2`: Diclofenac
- `3`: GiBleed
- `4`: NSAIDs

For the examples below, use `NSAIDs` (`cohortId = 4`) as the **target cohort**
and `Celecoxib`, `Diclofenac`, and `GiBleed` (`cohortId = 1:3`) as **event cohorts**.

## Run Post-Index Pathway Analysis

This is the default mode. It asks: what events happen after entry into the target cohort?

```{r post_index, eval=FALSE}
postIndexResults <- executeCohortPathways(
  connectionDetails = connectionDetails,
  cohortDatabaseSchema = "main",
  cohortTableName = "cohort",
  targetCohortIds = 4,
  eventCohortIds = c(1, 2, 3),
  maxDepth = 3,
  collapseWindow = 30
)
```

The result is a named list of analysis outputs:

```{r post_index_names, eval=FALSE}
names(postIndexResults)
```

The two most useful tables to inspect first are pathway-level counts and the
event-code mapping used to decode combinations:

```{r post_index_inspect, eval=FALSE}
head(postIndexResults$pathwaysAnalysisPathsData)
head(postIndexResults$pathwayAnalysisCodesLong)
```

## Run Pre-Index Pathway Analysis

Pre-index mode asks: what events occurred before the target cohort entry date,
within a configurable lookback window?

```{r pre_index, eval=FALSE}
preIndexResults <- executeCohortPathways(
  connectionDetails = connectionDetails,
  cohortDatabaseSchema = "main",
  cohortTableName = "cohort",
  targetCohortIds = 4,
  eventCohortIds = c(1, 2, 3),
  analysisType = "pre-index",
  lookbackStartDay = -365,
  lookbackEndDay = -1,
  maxDepth = 3,
  collapseWindow = 30
)
```

You can narrow the lookback window without changing any other arguments:

```{r pre_index_90, eval=FALSE}
preIndex90 <- executeCohortPathways(
  connectionDetails = connectionDetails,
  cohortDatabaseSchema = "main",
  cohortTableName = "cohort",
  targetCohortIds = 4,
  eventCohortIds = c(1, 2, 3),
  analysisType = "pre-index",
  lookbackStartDay = -90,
  lookbackEndDay = -1,
  maxDepth = 3
)
```

## Understanding the Returned Objects

`executeCohortPathways()` returns several tables, each serving a different purpose:

- `pathwayAnalysisStatsData`: summary-level analysis metadata and counts.
- `pathwaysAnalysisPathsData`: pathway sequences with `step1`, `step2`, ... and person counts.
- `pathwaysAnalysisEventsData`: event-level counts.
- `pathwaycomboIds`: unique event-combination codes observed in the pathways.
- `pathwayAnalysisCodesLong`: long-form decoding of combination codes into event cohorts.
- `isCombo`: identifies whether a code represents a single event or a multi-event combination.
- `pathwayAnalysisCodesData`: compact code lookup table.

For example, to inspect only the decoded event combinations:

```{r code_mapping, eval=FALSE}
subset(
  postIndexResults$pathwayAnalysisCodesLong,
  select = c(pathwayAnalysisGenerationId, code, targetCohortId, eventCohortId, isCombo, numberOfEvents)
)
```

## Using Separate Target and Event Cohort Tables

The core function also supports separate target and event tables. In the Eunomia
SQLite example, you can create those tables directly from `main.cohort`:

```{r separate_tables, eval=FALSE}
connection <- DatabaseConnector::connect(connectionDetails)

DatabaseConnector::executeSql(connection, "DROP TABLE IF EXISTS target_cohorts;")
DatabaseConnector::executeSql(connection, "DROP TABLE IF EXISTS event_cohorts;")

DatabaseConnector::executeSql(
  connection,
  "CREATE TABLE target_cohorts AS
     SELECT *
     FROM main.cohort
     WHERE cohort_definition_id = 4;"
)

DatabaseConnector::executeSql(
  connection,
  "CREATE TABLE event_cohorts AS
     SELECT *
     FROM main.cohort
     WHERE cohort_definition_id IN (1, 2, 3);"
)

resultsSeparateTables <- executeCohortPathways(
  connectionDetails = connectionDetails,
  cohortDatabaseSchema = "main",
  cohortTableName = "target_cohorts",
  outcomeDatabaseSchema = "main",
  outcomeTableName = "event_cohorts",
  targetCohortIds = 4,
  eventCohortIds = c(1, 2, 3)
)

DatabaseConnector::disconnect(connection)
```

This is useful when target cohorts and event cohorts are managed by different
ETL or cohort-generation steps.

## Building an Event Sequence Graph

The raw pathway output uses bitmask-encoded combo IDs. Use
`buildEventSequenceGraph()` to decode these into a directed **igraph** graph
with human-readable event names, transition edges, and probabilities.

**Note:** The simplified Eunomia example database produces only single-step
pathways (each patient has exactly one event after the target index date).
`buildEventSequenceGraph()` requires at least two steps to construct transition
edges.  On real-world OMOP data with richer treatment histories, the pathway
output from `executeCohortPathways()` will typically contain multiple steps and
can be passed directly to `buildEventSequenceGraph()`.

The example below constructs a small mock pathway result set that mirrors the
structure returned by `executeCohortPathways()`, so you can see the full
graph-building workflow in action:

```{r esg_mock_data, eval=FALSE}
# --- Mock cpResults with multi-step pathways ---
# Bitmask combo codes: 2 = Celecoxib, 4 = Diclofenac, 8 = GiBleed
mockPathsData <- data.frame(
  pathwayAnalysisGenerationId = rep(1L, 5),
  targetCohortId              = rep(4L, 5),
  step1      = c( 2L,  2L,  4L,  4L,  2L),
  step2      = c( 4L,  8L,  2L,  8L, NA),
  step3      = c( 8L, NA,   8L, NA,  NA),
  countValue = c(120L, 80L, 95L, 65L, 40L)
)

mockCodesLong <- data.frame(
  pathwayAnalysisGenerationId = rep(1L, 3),
  code           = c(2L, 4L, 8L),
  targetCohortId = rep(4L, 3),
  eventCohortId  = c(1L, 2L, 3L),
  isCombo        = rep(0L, 3),
  numberOfEvents = rep(1L, 3)
)

mockIsCombo <- data.frame(
  targetCohortId = rep(4L, 3),
  comboId        = c(2L, 4L, 8L),
  numberOfEvents = rep(1L, 3),
  isCombo        = rep(0L, 3)
)

mockCpResults <- list(
  pathwayAnalysisStatsData   = data.frame(
    pathwayAnalysisGenerationId = 1L,
    targetCohortId = 4L,
    countValue = 400L
  ),
  pathwaysAnalysisPathsData  = mockPathsData,
  pathwaysAnalysisEventsData = data.frame(eventCohortId = 1:3, countValue = c(240L, 215L, 360L)),
  pathwaycomboIds            = data.frame(comboIds = c(2L, 4L, 8L)),
  pathwayAnalysisCodesLong   = mockCodesLong,
  isCombo                    = mockIsCombo,
  pathwayAnalysisCodesData   = data.frame(
    pathwayAnalysisGenerationId = rep(1L, 3),
    code    = c(2L, 4L, 8L),
    isCombo = rep(0L, 3)
  )
)
```

Now build the graph using a generation set that maps cohort IDs to names:

```{r event_sequence_graph, eval=FALSE}
# Map cohort IDs to human-readable names
generationSet <- data.frame(
  cohortId   = c(1L, 2L, 3L, 4L),
  cohortName = c("Celecoxib", "Diclofenac", "GiBleed", "NSAIDs")
)

esg <- buildEventSequenceGraph(
  cpResults     = mockCpResults,
  generationSet = generationSet,
  maxSteps      = 3,
  minCount      = 1
)

# Print a summary
esg
```

When working with real `executeCohortPathways()` output, use the generation set
from `Eunomia::createCohorts()` (renaming `name` to `cohortName`):

```{r esg_real_note, eval=FALSE}
# With real data:
# generationSet <- Eunomia::createCohorts(connectionDetails)
# generationSet$cohortName <- generationSet$name
#
# esg <- buildEventSequenceGraph(
#   cpResults     = postIndexResults,
#   generationSet = generationSet,
#   maxSteps      = 3,
#   minCount      = 5
# )
```

The returned object is a list of class `"event_sequence_graph"` with four
components:

```{r graph_components, eval=FALSE}
# The igraph object — vertices are (event, step) pairs, edges are transitions
ig <- esg$graph

# Vertex attributes
igraph::V(ig)$eventName   # human-readable event names
igraph::V(ig)$step        # pathway step number
igraph::V(ig)$count       # patient count at this node
igraph::V(ig)$share       # share within the step (sums to 1)

# Edge attributes
igraph::E(ig)$weight      # patient count crossing this transition
igraph::E(ig)$probability # transition probability (sums to 1 per source)
igraph::E(ig)$sourceStep
igraph::E(ig)$targetStep

# Decoded pathways
head(esg$sequences)

# Summary statistics
esg$summary
```

### Quick visualization

`plot()` is defined on the returned object and produces a layered graph using
igraph's Sugiyama layout. Nodes are sized by patient count and colored by event
identity (same event = same color across steps). Edge widths are proportional
to transition weights.

```{r plot_esg, eval=FALSE}
# Default plot
plot(esg)

# Customized plot
plot(esg,
  colorPalette    = c("#1b9e77", "#d95f02", "#7570b3"),
  edgeWidthRange  = c(1, 10),
  vertexSizeRange = c(10, 30),
  main            = "Post-Index Treatment Pathways"
)
```

### Transition probabilities

Each edge carries a `probability` attribute — the fraction of patients at a
source event (within a step) who transition to each target:

```{r transition_probs, eval=FALSE}
# Extract edge data frame
edgeDf <- igraph::as_data_frame(esg$graph, what = "edges")

# View transitions from Step 1 to Step 2
edgeDf[edgeDf$sourceStep == 1, ]
```

### Downstream igraph analysis

Because the result is a standard igraph object, the full igraph API is
available for network analysis:

```{r igraph_analysis, eval=FALSE}
ig <- esg$graph

# Out-degree: how many distinct next-step events each node leads to
igraph::degree(ig, mode = "out")

# Weighted betweenness (inverse weight = lower traffic → higher betweenness)
igraph::betweenness(ig, weights = 1 / igraph::E(ig)$weight)

# Shortest weighted paths between all pairs
igraph::distances(ig, weights = 1 / igraph::E(ig)$weight)

# Identify hubs and authorities (HITS)
igraph::hub_score(ig, weights = igraph::E(ig)$weight)$vector
igraph::authority_score(ig, weights = igraph::E(ig)$weight)$vector

# Export to data frames for use outside igraph
vertDf <- igraph::as_data_frame(ig, what = "vertices")
edgeDf <- igraph::as_data_frame(ig, what = "edges")
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ExecuteCohortPathways.R
\name{executeCohortPathways}
\alias{executeCohortPathways}
\title{Execute cohort pathway analysis.}
\usage{
executeCohortPathways(
  connectionDetails = NULL,
  connection = NULL,
  cohortDatabaseSchema,
  cohortTableName = "cohort",
  outcomeDatabaseSchema = cohortDatabaseSchema,
  outcomeTableName = cohortTableName,
  tempEmulationSchema = getOption("sqlRenderTempEmulationSchema"),
  targetCohortIds,
  eventCohortIds,
  analysisType = "post-index",
  lookbackStartDay = -365,
  lookbackEndDay = -1,
  minCellCount = 5,
  allowRepeats = FALSE,
  maxDepth = 5,
  collapseWindow = 30
)
}
\arguments{
\item{connectionDetails}{An object of type connectionDetails as created using the
\code{\link[DatabaseConnector]{createConnectionDetails}} function in the
DatabaseConnector package. Can be left NULL if \code{connection} is
provided.}

\item{connection}{An object of type \code{connection} as created using the
\code{\link[DatabaseConnector]{connect}} function in the
DatabaseConnector package. Can be left NULL if connectionDetails
is provided, in which case a new connection will be opened at the start
of the function, and closed when the function finishes.}

\item{cohortDatabaseSchema}{Schema name where your \strong{target} cohort table resides.
For SQL Server this should include both database and schema,
e.g. 'scratch.dbo'.}

\item{cohortTableName}{The name of the \strong{target} cohort table (default: "cohort").}

\item{outcomeDatabaseSchema}{Schema name where your \strong{event/outcome} cohort table resides.
Defaults to \code{cohortDatabaseSchema} if not specified.}

\item{outcomeTableName}{The name of the \strong{event/outcome} cohort table.
Defaults to \code{cohortTableName} if not specified.}

\item{tempEmulationSchema}{Some database platforms like Oracle and Impala do not truly support
temp tables. To emulate temp tables, provide a schema with write
privileges where temp tables can be created.}

\item{targetCohortIds}{A vector of one or more Cohort Ids corresponding to target cohort(s).}

\item{eventCohortIds}{A vector of one or more Cohort Ids corresponding to event cohort(s).}

\item{analysisType}{Character. Either \code{"post-index"} (default) to analyze events
after the index date, or \code{"pre-index"} to analyze events
before the index date.}

\item{lookbackStartDay}{(Only used when \code{analysisType = "pre-index"}.)
Start of the lookback window in days relative to the index date
(default: -365). Must be negative.}

\item{lookbackEndDay}{(Only used when \code{analysisType = "pre-index"}.)
End of the lookback window in days relative to the index date
(default: -1). Must be negative.}

\item{minCellCount}{(Default = 5) The minimum cell count for fields containing person
counts or fractions.}

\item{allowRepeats}{(Default = FALSE) Allow cohort events/combos to appear multiple
times in the same pathway.}

\item{maxDepth}{(Default = 5) Maximum number of steps in a given pathway.}

\item{collapseWindow}{(Default = 30) Any dates found within the specified collapse days
will be reassigned the earliest date.}
}
\value{
\if{html}{\out{<div class="sourceCode">}}\preformatted{                 A list of data frames containing pathway analysis results.
}\if{html}{\out{</div>}}
}
\description{
Runs the cohort pathways on all instantiated combinations of target and event cohorts.
Assumes the cohorts have already been instantiated.

Supports two analysis types:
\itemize{
\item \code{"post-index"} (default): events occurring \strong{after} the target cohort
index date (within the target cohort period).
\item \code{"pre-index"}: events occurring \strong{before} the target cohort index date,
in a configurable lookback window. For example, what events occurred in
the 365 days before cohort entry.
}

Target and event cohorts can reside in different schemas/tables:
\itemize{
\item \code{cohortDatabaseSchema} / \code{cohortTableName} for target cohorts.
\item \code{outcomeDatabaseSchema} / \code{outcomeTableName} for event (outcome) cohorts.
If not specified, defaults to the same as the target.
}
}
\examples{
\donttest{
if (requireNamespace("Eunomia", quietly = TRUE) &&
    exists("createCohorts", where = asNamespace("Eunomia"), inherits = FALSE)) {
  connectionDetails <- Eunomia::getEunomiaConnectionDetails()
  createCohorts <- get("createCohorts", envir = asNamespace("Eunomia"))
  createCohorts(connectionDetails)

  postIndexResults <- executeCohortPathways(
    connectionDetails = connectionDetails,
    cohortDatabaseSchema = "main",
    cohortTableName = "cohort",
    targetCohortIds = 4,
    eventCohortIds = c(1, 2, 3),
    maxDepth = 3
  )

  preIndexResults <- executeCohortPathways(
    connectionDetails = connectionDetails,
    cohortDatabaseSchema = "main",
    cohortTableName = "cohort",
    targetCohortIds = 4,
    eventCohortIds = c(1, 2, 3),
    analysisType = "pre-index",
    lookbackStartDay = -365,
    lookbackEndDay = -1,
    maxDepth = 3
  )

  names(postIndexResults)
  names(preIndexResults)
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/EventSequenceGraph.R
\name{plot.event_sequence_graph}
\alias{plot.event_sequence_graph}
\title{Plot an event sequence graph.}
\usage{
\method{plot}{event_sequence_graph}(
  x,
  colorPalette = NULL,
  edgeWidthRange = c(0.5, 8),
  vertexSizeRange = c(8, 25),
  vertexLabelCex = 0.7,
  main = "Event Sequence Graph",
  ...
)
}
\arguments{
\item{x}{An \code{event_sequence_graph} object (returned by
\code{\link{buildEventSequenceGraph}}).}

\item{colorPalette}{Character vector of hex colors for nodes. If \code{NULL}
(default), a built-in palette is used. Colors are mapped by unique event
name (same event = same color across steps).}

\item{edgeWidthRange}{Numeric vector of length 2. Min and max edge widths
(default: \code{c(0.5, 8)}).}

\item{vertexSizeRange}{Numeric vector of length 2. Min and max vertex sizes
(default: \code{c(8, 25)}).}

\item{vertexLabelCex}{Label size multiplier (default: 0.7).}

\item{main}{Plot title (default: "Event Sequence Graph").}

\item{...}{Additional arguments passed to \code{\link[igraph]{plot.igraph}}.}
}
\value{
Invisibly returns the igraph object.
}
\description{
Produces a layered plot of the event sequence graph using \code{\link[igraph]{plot.igraph}}.
Nodes are laid out by step (left to right), with edge widths proportional to
patient counts and node sizes proportional to event counts within each step.
}
\examples{
if (requireNamespace("igraph", quietly = TRUE)) {
  cpResults <- list(
    pathwaysAnalysisPathsData = data.frame(
      pathwayAnalysisGenerationId = c(1L, 1L),
      targetCohortId = c(10L, 10L),
      step1 = c(2L, 4L),
      step2 = c(4L, NA_integer_),
      countValue = c(10L, 5L)
    ),
    pathwayAnalysisCodesLong = data.frame(
      pathwayAnalysisGenerationId = c(1L, 1L),
      code = c(2L, 4L),
      targetCohortId = c(10L, 10L),
      eventCohortId = c(1L, 2L),
      isCombo = c(0L, 0L),
      numberOfEvents = c(1L, 1L)
    ),
    isCombo = data.frame(
      targetCohortId = c(10L, 10L),
      comboId = c(2L, 4L),
      numberOfEvents = c(1L, 1L),
      isCombo = c(0L, 0L)
    )
  )

  generationSet <- data.frame(
    cohortId = c(1L, 2L, 10L),
    cohortName = c("Celecoxib", "Diclofenac", "NSAIDs")
  )

  esg <- buildEventSequenceGraph(cpResults, generationSet)
  plot(esg)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/OdysseusPathwayModule-package.R
\docType{package}
\name{OdysseusPathwayModule-package}
\alias{OdysseusPathwayModule}
\alias{OdysseusPathwayModule-package}
\title{OdysseusPathwayModule: Cohort Pathway Analysis with Pre-Index Event Support}
\description{
Provides cohort pathway analysis for Observational Medical Outcomes Partnership (OMOP) Common Data Model databases, including both standard (post-index) and pre-index pathway analyses. The pre-index analysis identifies sequences of events occurring in a lookback window before the target cohort index date. Built on the 'CohortPathways' analysis framework originally developed by Christopher Knoll and the Observational Health Data Sciences and Informatics community through 'WebAPI'. Methodological background and the originating implementation are described in \url{https://github.com/OHDSI/CohortPathways}.
}
\author{
\strong{Maintainer}: Alexander Alexeyuk \email{AlexanderAlexeyuk@gmail.com}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/EventSequenceGraph.R
\name{buildEventSequenceGraph}
\alias{buildEventSequenceGraph}
\title{Build an event sequence graph from pathway analysis results.}
\usage{
buildEventSequenceGraph(
  cpResults,
  generationSet,
  maxSteps = NULL,
  minCount = 1
)
}
\arguments{
\item{cpResults}{A list returned by \code{\link{executeCohortPathways}}.
Must contain at least \code{pathwaysAnalysisPathsData},
\code{pathwayAnalysisCodesLong}, and \code{isCombo}.}

\item{generationSet}{A data frame with at least two columns:
\code{cohortId} (integer) and \code{cohortName} (character).
Used to map event cohort IDs to descriptive names.}

\item{maxSteps}{(Optional) Maximum number of steps to include. If \code{NULL}
(default), all non-empty steps are used.}

\item{minCount}{(Default = 1) Minimum patient count for a pathway to be
included in the output.}
}
\value{
A list of class \code{"event_sequence_graph"} with components
\code{graph} (igraph object), \code{sequences}, \code{summary}, and
\code{eventNameLookup}.
}
\description{
Takes the output of \code{\link{executeCohortPathways}} and a cohort name
mapping, decodes the bitmask-encoded combo IDs into human-readable event
names, and builds a directed \pkg{igraph} graph representing event
transitions across pathway steps.

Nodes in the graph represent unique \emph{(event, step)} combinations.
Edges represent directed transitions between consecutive steps, weighted by
patient counts and annotated with transition probabilities.

The returned object includes:
\itemize{
\item \strong{graph}: A directed, weighted \code{\link[igraph]{igraph}}
object. Vertex attributes: \code{name}, \code{eventName},
\code{step}, \code{count}, \code{share}. Edge attributes:
\code{weight} (= patient count), \code{probability},
\code{sourceStep}, \code{targetStep}.
\item \strong{sequences}: The fully decoded pathway sequences with a
human-readable \code{pathway} string column.
\item \strong{summary}: High-level statistics (total pathways, patients,
depth, unique events, vertex/edge counts).
\item \strong{eventNameLookup}: Named character vector mapping combo IDs
to event names.
}
}
\examples{
if (requireNamespace("igraph", quietly = TRUE)) {
  cpResults <- list(
    pathwaysAnalysisPathsData = data.frame(
      pathwayAnalysisGenerationId = c(1L, 1L),
      targetCohortId = c(10L, 10L),
      step1 = c(2L, 4L),
      step2 = c(4L, NA_integer_),
      countValue = c(10L, 5L)
    ),
    pathwayAnalysisCodesLong = data.frame(
      pathwayAnalysisGenerationId = c(1L, 1L),
      code = c(2L, 4L),
      targetCohortId = c(10L, 10L),
      eventCohortId = c(1L, 2L),
      isCombo = c(0L, 0L),
      numberOfEvents = c(1L, 1L)
    ),
    isCombo = data.frame(
      targetCohortId = c(10L, 10L),
      comboId = c(2L, 4L),
      numberOfEvents = c(1L, 1L),
      isCombo = c(0L, 0L)
    )
  )

  generationSet <- data.frame(
    cohortId = c(1L, 2L, 10L),
    cohortName = c("Celecoxib", "Diclofenac", "NSAIDs")
  )

  esg <- buildEventSequenceGraph(cpResults, generationSet)
  esg$summary
  igraph::gorder(esg$graph)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/EventSequenceGraph.R
\name{.buildEventNameLookup}
\alias{.buildEventNameLookup}
\title{Build a named lookup vector: character(comboId) -> "EventA + EventB"}
\usage{
.buildEventNameLookup(codesLong, generationSet)
}
\description{
Build a named lookup vector: character(comboId) -> "EventA + EventB"
}
\keyword{internal}
